import json
import mimetypes
import shutil
from pathlib import Path
from typing import Any

import requests
import yaml

from pimpmyrice.config import (
    BASE_STYLE_FILE,
    CONFIG_FILE,
    JSON_SCHEMA_DIR,
    LOG_FILE,
    MODULES_DIR,
    PALETTES_DIR,
    PIMP_CONFIG_DIR,
    STYLES_DIR,
    TEMP_DIR,
    THEMES_DIR,
)
from pimpmyrice.keywords import default_base_style
from pimpmyrice.logger import get_logger
from pimpmyrice.utils import Result

log = get_logger(__name__)


def load_yaml(file: Path) -> dict[str, Any]:
    with open(file) as f:
        return dict(yaml.load(f, Loader=yaml.Loader))


def save_yaml(file: Path, data: dict[str, Any]) -> None:
    dump = yaml.dump(data, indent=4, default_flow_style=False)

    if file.name in ("module.yaml", "theme.yaml"):
        schema_str = f'# yaml-language-server: $schema=../../.json_schemas/{file.name.split(".")[0]}.json\n\n'
        dump = schema_str + dump

    with open(file, "w") as f:
        f.write(dump)


def load_json(file: Path) -> dict[str, Any]:
    with open(file) as f:

        data = json.load(f)
        data.pop("$schema", None)

        return data  # type: ignore


def save_json(file: Path, data: dict[str, Any]) -> None:
    if file.name in ("module.json", "theme.json") and file.parent != JSON_SCHEMA_DIR:
        data["$schema"] = f'../../.json_schemas/{file.name.split(".")[0]}.json'

    dump = json.dumps(data, indent=4)

    with open(file, "w") as f:
        f.write(dump)


def import_image(image_path: Path, theme_dir: Path) -> Path:
    if (
        not image_path.exists() or not image_path.is_file()
    ):  # to do: process files/folders
        raise FileNotFoundError(f'file not found at "{image_path}"')

    dest = theme_dir / image_path.name
    if (dest).exists():
        raise Exception(f'file already exists at "{dest}"')

    shutil.copy(image_path, theme_dir)
    log.debug(f'image "{image_path.name}" imported')
    return dest


def check_config_dirs() -> None:
    for dir in [
        THEMES_DIR,
        STYLES_DIR,
        PALETTES_DIR,
        MODULES_DIR,
        TEMP_DIR,
        JSON_SCHEMA_DIR,
    ]:
        dir.mkdir(exist_ok=True, parents=True)

    if not BASE_STYLE_FILE.exists():
        save_json(BASE_STYLE_FILE, default_base_style)
    if not CONFIG_FILE.exists():
        config = {"theme": None, "mode": "dark"}
        save_json(CONFIG_FILE, config)
    if not LOG_FILE.exists():
        LOG_FILE.touch()

    # if not VENV_DIR.exists():
    #     create_venv()


def download_file(url: str, destination: Path = TEMP_DIR) -> Result[Path]:
    # TODO better filename

    res = Result[Path]()

    try:
        response = requests.get(url, stream=True)

        if response.status_code != 200:
            return res.error(
                f"Failed to download image. Status code: {response.status_code}"
            )

        content_type = response.headers.get("content-type")
        file_extension = (
            mimetypes.guess_extension(content_type) if content_type else None
        )

        if not file_extension:
            file_extension = ".jpg"

        filename = url.split("/")[-1].split("?")[0]

        if not filename.endswith(file_extension):
            filename = filename + file_extension

        save_path = destination / filename

        tries = 1
        while save_path.exists():
            save_path = (
                save_path.parent / f"{save_path.stem}_{tries+1}{save_path.suffix}"
            )

        with open(save_path, "wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

        res.value = save_path
        return res

    except Exception as e:
        return res.exception(e)



from __future__ import annotations

import os
import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Generator

from pydantic import BaseModel, create_model

from pimpmyrice.config import HOME_DIR, JSON_SCHEMA_DIR
from pimpmyrice.doc import __doc__ as cli_doc
from pimpmyrice.files import save_json
from pimpmyrice.logger import get_logger
from pimpmyrice.module_utils import Module
from pimpmyrice.theme_utils import Theme
from pimpmyrice.utils import Result

if TYPE_CHECKING:
    from pimpmyrice.theme import ThemeManager

log = get_logger(__name__)


def create_dynamic_model(name: str, source: dict[str, Any]) -> BaseModel:
    fields: dict[str, Any] = {}
    for key, value in source.items():
        if isinstance(value, dict):
            nested_model = create_dynamic_model(f"{name}_{key}", value)
            fields[key] = (nested_model, {})
        else:
            fields[key] = (type(value), value)

    model: BaseModel = create_model(name, **fields)

    return model


def generate_theme_json_schema(tm: ThemeManager) -> Result:
    res = Result()

    base_style = deepcopy(tm.base_style)

    for module in tm.mm.modules:
        if module not in base_style["modules_styles"]:
            base_style["modules_styles"][module] = {}

    style_model = create_dynamic_model("Style", base_style)
    style_schema = style_model.model_json_schema()

    theme_schema = Theme.model_json_schema()

    tags_schema = {
        "default": [],
        "items": {
            "anyOf": [
                {"type": "string"},
                {
                    "const": "",
                    "enum": list(tm.tags),
                    "type": "string",
                },
            ]
        },
        "title": "Tags",
        "type": "array",
        "uniqueItems": True,
    }

    theme_schema["properties"]["tags"] = tags_schema

    theme_schema["$defs"] = {**theme_schema["$defs"], **style_schema["$defs"]}
    style_schema.pop("$defs")
    style_schema["required"] = []

    theme_schema["$defs"]["Style"] = style_schema

    theme_schema["$defs"]["Mode"]["properties"]["style"] = {"$ref": "#/$defs/Style"}

    theme_schema["properties"]["style"] = {"$ref": "#/$defs/Style"}

    theme_schema["properties"].pop("name")
    theme_schema["properties"].pop("path")
    theme_schema["required"].remove("name")
    theme_schema["required"].remove("path")

    schema_path = JSON_SCHEMA_DIR / "theme.json"
    save_json(schema_path, theme_schema)

    return res.debug(f'theme schema saved to "{schema_path}"')


def generate_module_json_schema() -> Result:
    res = Result()

    module_schema = Module.model_json_schema()

    schema_path = JSON_SCHEMA_DIR / "module.json"
    save_json(schema_path, module_schema)

    return res.debug(f'module schema saved to "{schema_path}"')



import random
import shutil
from pathlib import Path
from typing import Any

import rich

from pimpmyrice import parsers, schemas
from pimpmyrice import theme_utils as tutils
from pimpmyrice.colors import (
    GlobalPalette,
    LinkPalette,
    Palette,
    exp_gen_palette,
    get_palettes,
)
from pimpmyrice.completions import generate_shell_suggestions
from pimpmyrice.config import BASE_STYLE_FILE, CONFIG_FILE, STYLES_DIR, THEMES_DIR
from pimpmyrice.events import EventHandler
from pimpmyrice.files import (
    check_config_dirs,
    download_file,
    load_json,
    save_json,
    save_yaml,
)
from pimpmyrice.logger import get_logger
from pimpmyrice.module import ModuleManager
from pimpmyrice.theme_utils import Mode, Style, Theme, ThemeConfig
from pimpmyrice.utils import AttrDict, Result, Timer

log = get_logger(__name__)


class ThemeManager:
    def __init__(self) -> None:
        timer = Timer()
        check_config_dirs()
        self.base_style = self.get_base_style()
        self.styles = self.get_styles()
        self.palettes = self.get_palettes()
        self.tags: set[str] = set()
        self.themes = self.get_themes()
        self.config = self.get_config()
        self.event_handler = EventHandler()
        self.mm = ModuleManager()

        # TODO move
        try:
            schemas.generate_theme_json_schema(self)
            schemas.generate_module_json_schema()
            generate_shell_suggestions(self)
        except Exception as e:
            log.exception(e)
            log.error("failed to generate suggestions")

        log.debug(f"ThemeManager initialized in {timer.elapsed():.4f} sec")

    def get_config(self) -> ThemeConfig:
        config = ThemeConfig(**load_json(CONFIG_FILE))
        if config.theme not in self.themes:
            config.theme = None
        return config

    def save_config(self) -> None:
        save_json(CONFIG_FILE, vars(self.config))

    @staticmethod
    def get_base_style() -> dict[str, Any]:
        try:
            return load_json(BASE_STYLE_FILE)
        except Exception:
            log.error("failed loading base_style.json")
            raise

    def save_base_style(self, base_style: dict[str, Any]) -> None:
        save_json(BASE_STYLE_FILE, base_style)
        self.base_style = base_style
        schemas.generate_theme_json_schema(self)

    @staticmethod
    def get_styles() -> dict[str, Style]:
        styles: dict[str, Style] = {f.stem: load_json(f) for f in STYLES_DIR.iterdir()}
        return styles

    @staticmethod
    def get_palettes() -> dict[str, GlobalPalette]:
        return get_palettes()

    def parse_theme(self, theme_path: Path) -> Theme:
        theme = parsers.parse_theme(theme_path, self.styles, self.palettes)

        return theme

    def get_themes(self) -> dict[str, Theme]:
        timer = Timer()

        themes: dict[str, Theme] = {}

        for directory in THEMES_DIR.iterdir():
            if not (directory / "theme.json").is_file():
                continue

            try:
                theme = self.parse_theme(directory)
            except Exception as e:
                log.exception(e)
                log.error(f'Error parsing theme "{directory.name}": {str(e)}')
                continue

            themes[directory.name] = theme

            for tag in theme.tags:
                self.tags.add(tag)

        log.debug(f"{len(themes)} themes loaded in {timer.elapsed():.4f} sec")

        return themes

    async def generate_theme(
        self,
        image: str,
        name: str | None = None,
        tags: set[str] | None = None,
        apply: bool = False,
    ) -> Result:
        res = Result()

        if image.startswith(("http://", "https://")):
            download_res = download_file(image)
            if download_res.value is None:
                res += download_res
                return res.error("could not download file")
            file = download_res.value
            res.info(f'downloaded "{file.name}"')
        else:
            file = Path(image)

        gen_res = await tutils.gen_from_img(image=file, name=name, themes=self.themes)
        res += gen_res
        if not gen_res.value:
            return res.error("could not generate theme")

        if tags:
            gen_res.value.tags = tags

        # TODO generate name here
        save_res = await self.save_theme(gen_res.value)
        res += save_res
        if save_res.value:
            res.success(f'theme "{save_res.value}" generated')
        else:
            return res.error("could not generate theme")

        if apply:
            apply_res = await self.apply_theme(save_res.value)
            res += apply_res

        res.ok = True
        return res

    async def rename_theme(
        self,
        theme_name: str,
        new_name: str,
    ) -> Result:
        res = Result()

        if theme_name not in self.themes:
            return res.error(f'theme "{theme_name}" not found')

        theme = self.themes[theme_name]
        old_name = theme.name
        theme.name = new_name

        save_res = await self.save_theme(theme, old_name=old_name)
        res += save_res

        if not save_res.value:
            return res.error(f'failed renaming theme "{theme_name}"')

        res.ok = True
        return res.success(f'renamed theme "{theme_name}" to "{new_name}"')

    async def save_theme(
        self,
        theme: Theme,
        old_name: str | None = None,
    ) -> Result[str]:
        res: Result[str] = Result()

        if not old_name:
            theme.name = tutils.valid_theme_name(name=theme.name, themes=self.themes)
            theme_dir = THEMES_DIR / theme.name
            theme_dir.mkdir()

        elif old_name != theme.name:
            theme.name = tutils.valid_theme_name(name=theme.name, themes=self.themes)
            theme_dir = THEMES_DIR / theme.name
            (THEMES_DIR / old_name).rename(theme_dir)
        else:
            theme_dir = THEMES_DIR / theme.name

        # NOTE full path update on rename is handled by dump_theme
        #      as it leaves only the filename
        theme.wallpaper.path = tutils.import_image(theme.wallpaper.path, theme_dir)
        for mode in theme.modes.values():
            if mode.wallpaper:
                mode.wallpaper.path = tutils.import_image(
                    mode.wallpaper.path, theme_dir
                )

        dump = tutils.dump_theme_for_file(theme)
        save_json(theme_dir / "theme.json", dump)
        # save_yaml(theme_dir / "theme.yaml", dump)

        parsed_theme = self.parse_theme(THEMES_DIR / theme.name)

        self.themes[theme.name] = parsed_theme

        if old_name and old_name != theme.name:
            self.themes.pop(old_name)

            if self.config.theme == old_name:
                self.config.theme = theme.name

        res.value = theme.name
        res.ok = True
        return res

    async def rewrite_themes(
        self,
        regen_colors: bool = False,
        name_includes: str | None = None,
        include_tags: set[str] | None = None,
        exclude_tags: set[str] | None = None,
    ) -> Result:
        res = Result()

        for theme in self.themes.values():
            if name_includes and name_includes not in theme.name:
                continue

            if include_tags and not any(tag in include_tags for tag in theme.tags):
                continue

            if exclude_tags and any(tag in exclude_tags for tag in theme.tags):
                continue

            mode_names = {"light", "dark", *theme.modes.keys()}

            if regen_colors:
                for mode_name in mode_names:
                    if mode_name not in theme.modes:
                        palette = await exp_gen_palette(
                            img=theme.wallpaper.path, light=("light" in mode_name)
                        )
                        theme.modes[mode_name] = Mode(
                            name=mode_name,
                            wallpaper=theme.wallpaper,
                            palette=palette,
                        )
                    else:
                        mode = theme.modes[mode_name]
                        if mode.wallpaper:
                            mode.palette = await exp_gen_palette(
                                img=mode.wallpaper.path, light=("light" in mode.name)
                            )
            save_res = await self.save_theme(theme=theme, old_name=theme.name)
            if save_res.value:
                res.success(f'theme "{theme.name}" rewritten')
            else:
                res += save_res

        res.ok = True
        return res

    def delete_theme(self, theme_name: str) -> Result:
        res = Result()

        if theme_name not in self.themes:
            return res.error(f'theme "{theme_name}" not found')

        theme = self.themes[theme_name]

        if not str(theme.path).startswith(str(THEMES_DIR)) or theme.path == THEMES_DIR:
            return res.error(f'"{theme.path}" not in "{THEMES_DIR}"')

        shutil.rmtree(theme.path)

        if theme_name == self.config.theme:
            self.config.theme = None
        self.themes.pop(theme_name)

        res.ok = True
        return res.success(f'theme "{theme_name}" deleted')

    async def apply_theme(
        self,
        theme_name: str | None = None,
        mode_name: str | None = None,
        palette_name: str | None = None,
        styles_names: list[str] | None = None,
        include_modules: list[str] | None = None,
        exclude_modules: list[str] | None = None,
        print_theme_dict: bool = False,
    ) -> Result:
        res = Result()

        if not theme_name:
            if not self.config.theme:
                return res.error("No current theme")
            theme_name = self.config.theme
        elif theme_name not in self.themes:
            return res.error(f'"{theme_name}" not found')

        if not mode_name:
            mode_name = self.config.mode

        r = tutils.gen_theme_dict(
            self,
            theme_name=theme_name,
            mode_name=mode_name,
            styles_names=styles_names,
            palette_name=palette_name,
        )
        res += r

        if not r.value:
            return res.error(
                f'error generating the theme_dict for theme "{theme_name}"'
            )

        theme_dict = r.value

        if print_theme_dict:
            pretty = rich.pretty.pretty_repr(theme_dict)
            res.info("generated theme_dict:\r\n" + pretty)

        res.info(f'applying theme "{theme_name}"...')

        modules_res = await self.mm.run_modules(
            theme_dict, include_modules, exclude_modules
        )

        res += modules_res

        self.config.theme = theme_name
        self.config.mode = mode_name
        self.save_config()

        # display_color_palette(palette.term)
        if res.errors:
            res.warning(f'theme "{theme_name}" {mode_name} applied with errors')
        else:
            res.success(f'theme "{theme_name}" {mode_name} applied')

        await self.event_handler.publish("theme_applied")

        res.ok = True
        return res

    async def set_random_theme(
        self,
        mode_name: str | None = None,
        styles_names: list[str] | None = None,
        palette_name: str | None = None,
        name_includes: str | None = None,
        include_modules: list[str] | None = None,
        exclude_modules: list[str] | None = None,
        include_tags: set[str] | None = None,
        exclude_tags: set[str] | None = None,
        print_theme_dict: bool = False,
    ) -> Result:
        res = Result()

        themes_list: list[Theme] = []

        for theme in self.themes.values():
            if theme.name == self.config.theme:
                continue

            if name_includes and name_includes not in theme.name:
                continue

            if include_tags and not any(tag in include_tags for tag in theme.tags):
                continue

            if exclude_tags and any(tag in exclude_tags for tag in theme.tags):
                continue

            themes_list.append(theme)

        if len(themes_list) < 1:
            return res.error("no theme found")

        theme_name = random.choice(themes_list).name
        apply_res = await self.apply_theme(
            theme_name,
            mode_name=mode_name,
            styles_names=styles_names,
            palette_name=palette_name,
            include_modules=include_modules,
            exclude_modules=exclude_modules,
            print_theme_dict=print_theme_dict,
        )

        res += apply_res

        res.ok = True
        return res

    async def toggle_mode(self) -> Result:
        if not self.config.theme:
            return Result().error("no theme set")

        mode_name = "light" if self.config.mode == "dark" else "dark"

        return await self.apply_theme(mode_name=mode_name)

    async def set_mode(self, mode_name: str) -> Result:
        if not self.config.theme:
            return Result().error("no theme set")

        return await self.apply_theme(mode_name=mode_name)

    async def add_tags(self, themes_names: list[str], tags: set[str]) -> Result:
        res = Result()

        for theme_name in themes_names:
            if theme_name not in self.themes:
                res.error('theme "{theme_name}" not found')
                continue

            theme = self.themes[theme_name]

            for tag in tags:
                theme.tags.add(tag)
                save_res = await self.save_theme(theme, theme.name)
                res += save_res
                if not save_res.errors:
                    res.info(f'tag "{tag}" added to theme "{theme.name}"')

        res.ok = True
        return res

    async def remove_tags(self, themes_names: list[str], tags: set[str]) -> Result:
        res = Result()

        if len(themes_names) == 0:
            themes_names = list(self.themes.keys())

        for theme_name in themes_names:
            if theme_name not in self.themes:
                res.error('theme "{theme_name}" not found')
                continue

            theme = self.themes[theme_name]

            for tag in tags:
                if tag in theme.tags:
                    theme.tags.remove(tag)
                    save_res = await self.save_theme(theme, theme.name)
                    res += save_res
                    if not save_res.errors:
                        res.info(f'tag "{tag}" removed from theme "{theme.name}"')

        res.ok = True
        return res

    async def list_themes(self) -> Result:
        res = Result()

        res.info("\nNAME\t\t\tTAGS\n")
        for theme in self.themes.values():
            res.info(f"{theme.name:10}\t\t{', '.join(theme.tags)}")

        res.ok = True
        return res

    async def list_tags(self) -> Result:
        res = Result()

        res.info("\n".join(self.tags))

        res.ok = True
        return res

    async def list_palettes(self) -> Result:
        res = Result()

        for palette in self.palettes:
            res.info(f"{palette}")

        res.ok = True
        return res

    async def list_styles(self) -> Result:
        res = Result()

        for style in self.styles:
            res.info(f"{style}")

        res.ok = True
        return res

    async def export_theme(
        self,
        theme_name: str,
        out_dir: Path,
        mode_name: str | None = None,
        palette_name: str | None = None,
        styles_names: list[str] | None = None,
        include_modules: list[str] | None = None,
        exclude_modules: list[str] | None = None,
        print_theme_dict: bool = False,
    ) -> Result:

        res = Result()

        if theme_name not in self.themes:
            return res.error(f'theme "{theme_name}" not found')

        if not mode_name:
            mode_name = self.config.mode

        dump_dir = out_dir / f"{theme_name}_{mode_name}"

        if dump_dir.exists():
            return res.error(f'directory "{dump_dir}" already exists')

        gen_res = tutils.gen_theme_dict(
            self,
            theme_name=theme_name,
            mode_name=mode_name,
            styles_names=styles_names,
            palette_name=palette_name,
        )
        res += gen_res

        if not gen_res.value:
            return res.error(
                f'error generating the theme_dict for theme "{theme_name}"'
            )
        theme_dict = gen_res.value

        if print_theme_dict:
            pretty = rich.pretty.pretty_repr(theme_dict)
            res.info("generated theme_dict:\r\n" + pretty)

        modules_res = await self.mm.run_modules(
            theme_dict, include_modules, exclude_modules, dump_dir
        )

        res += modules_res
        if not modules_res.value:
            return res.error(f'error exporting theme "{theme_name}"')

        theme = self.themes[theme_name]

        wp = theme.modes[mode_name].wallpaper
        if not wp:
            return res

        shutil.copy(wp.path, dump_dir)

        readme = f"""# "{theme_name}" {mode_name} theme dotfiles

Dump generated with [pimp](https://github.com/daddodev/pimpmyrice) `export theme`

## Requirements:

"""

        for module_name in modules_res.value:
            readme += f"- {module_name}\n"

        with open(dump_dir / "README.md", "w") as f:
            f.write(readme)

        res.success(f'theme "{theme_name}" exported to {dump_dir}')

        res.ok = True
        return res



from __future__ import annotations

import os
import time
from collections.abc import MutableMapping
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generic

import cv2
import jinja2
import psutil
from typing_extensions import TypeVar

from pimpmyrice.config import CONFIG_DIR, HOME_DIR, MODULES_DIR
from pimpmyrice.logger import LogLevel, get_logger

# import jinja2schema


log = get_logger(__name__)


@dataclass
class ResultRecord:
    msg: str
    level: LogLevel

    def dump(self) -> dict[str, str]:
        d = {"msg": self.msg, "level": self.level.name}
        return d


T = TypeVar("T", default=None)


@dataclass
class Result(Generic[T]):
    value: T | None = None
    name: str | None = None
    errors: int = 0
    records: list[ResultRecord] = field(default_factory=list)
    time: float = 0
    ok: bool = False

    def __log(self, record: ResultRecord, name: str | None = None) -> None:
        if name:
            record.msg = f"{name}: {record.msg}"

        if log.isEnabledFor(record.level.value):
            log._log(level=record.level.value, msg=record.msg, args={})
        self.records.append(record)

    def debug(self, msg: str, name: str | None = None) -> Result[T]:
        self.__log(ResultRecord(msg, LogLevel.DEBUG), name)
        return self

    def info(self, msg: str, name: str | None = None) -> Result[T]:
        self.__log(ResultRecord(msg, LogLevel.INFO), name)
        return self

    def success(self, msg: str, name: str | None = None) -> Result[T]:
        self.__log(ResultRecord(msg, LogLevel.SUCCESS), name)
        return self

    def warning(self, msg: str, name: str | None = None) -> Result[T]:
        self.__log(ResultRecord(msg, LogLevel.WARNING), name)
        return self

    def error(self, msg: str, name: str | None = None) -> Result[T]:
        self.__log(ResultRecord(msg, LogLevel.ERROR), name)
        self.errors += 1
        return self

    def exception(
        self, exception: Exception, message: str | None = None, name: str | None = None
    ) -> Result[T]:
        exc_str = str(exception).strip()

        if message:
            msg = f"{message}:\r\n{' '*(len(name)+2 if name else 0)}{exc_str}"
        else:
            msg = exc_str

        self.__log(ResultRecord(msg, LogLevel.ERROR), name)

        self.errors += 1

        if log.isEnabledFor(LogLevel.DEBUG.value):
            log.exception(exception)

        return self

    def __add__(self, other: Result[Any]) -> Result[T]:
        r: Result[T] = Result(
            name=self.name,
            value=self.value,
            errors=self.errors + other.errors,
            records=[*self.records, *other.records],
        )
        return r

    def dump(self) -> dict[str, Any]:
        d = deepcopy(vars(self))
        d["records"] = [r.dump() for r in self.records]
        return d


class Timer:
    def __init__(self) -> None:
        self.start = time.perf_counter()

    def elapsed(self) -> float:
        return time.perf_counter() - self.start


# def get_template_keywords(template: str) -> list[str]:


class AttrDict(dict[str, Any]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.__dict__ = self
        super().__init__(*args, **kwargs)

        for k in self:
            if isinstance(self[k], dict):
                self[k] = AttrDict(self[k])

    def __setitem__(self, key: Any, value: Any) -> None:
        if isinstance(value, dict):
            value = AttrDict(value)
        super().__setitem__(key, value)

    def __add__(self, other: DictOrAttrDict) -> AttrDict:
        def merged(base: AttrDict, to_add: AttrDict) -> AttrDict:
            base = deepcopy(base)
            to_add = deepcopy(to_add)
            for k, v in to_add.items():
                if isinstance(v, (dict, AttrDict)):
                    if k in base:
                        base[k] = merged(base[k], to_add[k])
                    else:
                        base[k] = to_add[k]
                else:
                    base[k] = v
            return base

        return merged(self, AttrDict(other))


DictOrAttrDict = TypeVar("DictOrAttrDict", dict[str, Any], AttrDict)


def process_template(template: str, values: dict[str, Any]) -> str:
    # get_template_keywords(template)
    templ = jinja2.Environment(undefined=jinja2.StrictUndefined).from_string(template)
    rendered: str = templ.render(**values)
    return rendered


def parse_string_vars(
    string: str,
    theme_dict: dict[str, Any] | None = None,
    module_name: str | None = None,
) -> str:
    # TODO capitalize

    d = {"home_dir": HOME_DIR, "config_dir": CONFIG_DIR}
    if module_name:
        d["module_dir"] = MODULES_DIR / module_name
        d["templates_dir"] = MODULES_DIR / module_name / "templates"
        d["files_dir"] = MODULES_DIR / module_name / "files"
    if not theme_dict:
        theme_dict = d
    else:
        theme_dict |= d
    res = process_template(string, theme_dict)
    expanded: str = os.path.expanduser(res)
    return expanded


def is_process_running(name: str | None = None, pid: int | None = None) -> bool:
    if (not name and not pid) or (name and pid):
        raise Exception("provide either process pid or name")

    attr = "name" if name else "pid"
    val = name if name else pid

    for proc in psutil.process_iter([attr]):
        if proc.info[attr] == val:
            return True
    return False


def is_locked(lockfile: Path) -> tuple[bool, int]:
    if lockfile.exists():
        with open(lockfile, "r") as f:
            file_pid = int(f.read())

        if is_process_running(pid=file_pid):
            return True, file_pid

        lockfile.unlink()
    return False, 0


class Lock:
    def __init__(self, lockfile: Path) -> None:
        self.lockfile = lockfile

    def __enter__(self) -> None:
        pid = os.getpid()
        with open(self.lockfile, "w") as f:
            f.write(str(pid))

    def __exit__(self, *_: Any) -> None:
        self.lockfile.unlink()


def get_thumbnail(image_path: Path, max_px: int = 1024) -> Path:
    thumb_path = (
        image_path.parent / f".{image_path.stem}_thumb_{max_px}{image_path.suffix}"
    )

    if thumb_path.is_file():
        return thumb_path

    img = cv2.imread(str(image_path))
    fx = max_px / img.shape[1]
    fy = max_px / img.shape[0]
    f = min(fx, fy)
    size = (int(img.shape[1] * f), int(img.shape[0] * f))
    resized = cv2.resize(img, size)

    cv2.imwrite(str(thumb_path), resized)

    log.debug(f'{max_px}x{max_px} thumbnail generated for "{image_path}"')

    return thumb_path



import logging
from enum import Enum
from typing import Any

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

SUCCESS_LEVEL_NUM = 25


class LogLevel(Enum):
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    SUCCESS = SUCCESS_LEVEL_NUM
    INFO = logging.INFO
    DEBUG = logging.DEBUG


class Logger(logging.Logger):
    def success(self, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(SUCCESS_LEVEL_NUM):
            self._log(SUCCESS_LEVEL_NUM, message, args, **kwargs)


def set_up_logging() -> None:
    logging.setLoggerClass(Logger)
    logging.addLevelName(SUCCESS_LEVEL_NUM, "SUCCESS")
    console = Console(theme=Theme({"logging.level.success": "green"}))
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        datefmt="[X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True,
                tracebacks_show_locals=False,
                console=console,
                show_path=False,
                show_time=False,
                # level=logging.INFO,
            )
        ],
    )


def get_logger(name: str) -> Logger:
    log = logging.getLogger(name)
    return log  # type: ignore


# def to_file(msg: str) -> None:
#     with open(LOG_FILE, "a") as f:
#         f.write(msg + "\n")
#
#
# def format_msg(name: str, msg: str, level: LogLevel) -> str:
#     colors = {
#         LogLevel.ERROR: 1,
#         LogLevel.WARNING: 3,
#         LogLevel.SUCCESS: 2,
#         LogLevel.INFO: 7,
#         LogLevel.DEBUG: 6,
#     }
#
#     virtual_term = False
#     try:
#         term_size = os.get_terminal_size().columns
#     except Exception:
#         virtual_term = True
#         term_size = 80
#
#     msg_lines = [
#         f"{line:<9}" if i > 0 else line for i, line in enumerate(msg.splitlines())
#     ]
#     # origin_length = term_size - len(msg_lines[-1])
#     # if len(msg_lines) == 1:
#     #     origin_length -= 9
#     # if origin_length < len(name) + 1:
#     #     origin_length = 0
#     #     name = ""
#     msg = "\n".join(msg_lines)
#     # formatted = f"{level.name:<9}{msg}{name:>{origin_length}}"
#     formatted = f"{level.name:<9}{msg}"
#     if not virtual_term:
#         formatted = f"\033[3{colors[level]}m{formatted}\033[00m"
#
#     return formatted



from __future__ import annotations

import string
import unicodedata
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ItemsView,
    KeysView,
    Literal,
    Tuple,
    TypedDict,
    TypeVar,
    ValuesView,
)

from pydantic import BaseModel, Field, computed_field
from pydantic.json_schema import SkipJsonSchema

from pimpmyrice import files
from pimpmyrice.colors import Color, LinkPalette, Palette, exp_gen_palette
from pimpmyrice.config import JSON_SCHEMA_DIR
from pimpmyrice.logger import get_logger
from pimpmyrice.module_utils import FileAction, Module
from pimpmyrice.utils import AttrDict, DictOrAttrDict, Result, get_thumbnail

if TYPE_CHECKING:
    from pimpmyrice.theme import ThemeManager

log = get_logger(__name__)


Style = dict[str, Any]


class ThemeConfig(BaseModel):
    theme: str | None = None
    mode: str = "dark"


class Mode(BaseModel):
    name: SkipJsonSchema[str] = Field(exclude=True)
    palette: LinkPalette | Palette
    wallpaper: Wallpaper | None = None
    style: Style = {}


class WallpaperMode(str, Enum):
    FILL = "fill"
    FIT = "fit"

    def __str__(self) -> str:
        return self.value


class Wallpaper(BaseModel):
    path: Path
    mode: WallpaperMode = WallpaperMode.FILL

    @computed_field  # type: ignore
    @property
    def thumb(self) -> Path:
        t = get_thumbnail(self.path)
        return t


class Theme(BaseModel):
    path: Path = Field()
    name: str = Field()
    wallpaper: Wallpaper
    modes: dict[str, Mode] = {}
    style: Style = {}
    tags: set[str] = set()


def dump_theme_for_file(theme: Theme) -> dict[str, Any]:
    dump = theme.model_dump(
        mode="json",
        exclude={
            "name": True,
            "path": True,
            "wallpaper": {"thumb"},
            "modes": {"__all__": {"wallpaper": {"thumb"}}},
        },
    )

    for mode in dump["modes"].values():
        if not mode["style"]:
            mode.pop("style")

        if mode["wallpaper"] == dump["wallpaper"]:
            mode.pop("wallpaper")
        else:
            mode["wallpaper"]["path"] = str(Path(mode["wallpaper"]["path"]).name)
            if mode["wallpaper"]["mode"] == "fill":
                mode["wallpaper"].pop("mode")

    if not dump["style"]:
        dump.pop("style")

    dump["wallpaper"]["path"] = str(Path(dump["wallpaper"]["path"]).name)

    if dump["wallpaper"]["mode"] == "fill":
        dump["wallpaper"].pop("mode")

    if not dump["tags"]:
        dump.pop("tags")

    # print("dump for file:", json.dumps(dump, indent=4))
    return dump


async def gen_from_img(
    image: Path,
    themes: dict[str, Theme],
    name: str | None = None,
) -> Result[Theme]:
    res: Result[Theme] = Result()

    if not image.is_file():
        return res.error(f'image not found at "{image}"')

    dark_colors = await exp_gen_palette(image)
    light_colors = await exp_gen_palette(image, light=True)
    modes = {
        "dark": Mode(name="dark", wallpaper=Wallpaper(path=image), palette=dark_colors),
        "light": Mode(
            name="light", wallpaper=Wallpaper(path=image), palette=light_colors
        ),
    }

    theme_name = valid_theme_name(name or image.stem, themes)
    theme = Theme(
        name=theme_name, path=Path(), wallpaper=Wallpaper(path=image), modes=modes
    )

    res.value = theme

    return res


def resolve_refs(
    data: DictOrAttrDict, theme_dict: DictOrAttrDict | None = None
) -> Tuple[DictOrAttrDict, list[str]]:
    if not theme_dict:
        theme_dict = deepcopy(data)

    unresolved = []

    for key, value in data.items():
        if isinstance(value, dict):
            data[key], pending = resolve_refs(value, theme_dict)
            for p in pending:
                unresolved.append(f"{key}.{p}")
        elif isinstance(value, str) and value.startswith("$"):
            ref = value[1:]

            ref_slices = ref.split(".")
            d = theme_dict

            while len(ref_slices) > 1:
                if ref_slices[0] in d:
                    d = d[ref_slices[0]]
                else:
                    unresolved.append(key)
                    break
                ref_slices.pop(0)

            if isinstance(d, Color):
                res = getattr(d, ref_slices[0])
                data[key] = Color(res)
            elif ref_slices[0] in d and not str(d[ref_slices[0]]).startswith("$"):
                data[key] = d[ref_slices[0]]
            else:
                unresolved.append(key)

    return data, unresolved


def gen_theme_dict(
    tm: ThemeManager,
    theme_name: str,
    mode_name: str,
    palette_name: str | None = None,
    styles_names: list[str] | None = None,
) -> Result[AttrDict]:

    res: Result[AttrDict] = Result()

    theme = tm.themes[theme_name]

    if mode_name not in theme.modes:
        new_mode = [*theme.modes.keys()][0]
        res.warning(f'"{mode_name}" mode not present in theme, applying "{new_mode}"')
        mode_name = new_mode

    styles: list[Style] = []

    if theme.style:
        if from_global := theme.style.get("from_global"):
            if from_global not in tm.styles:
                return res.error(
                    f'global style "{from_global}" not found in {list(tm.styles)}'
                )
            theme_style = AttrDict(**tm.styles[from_global]) + theme.style
            styles.append(theme_style)
        else:
            styles.append(theme.style)

    if mode_style := theme.modes[mode_name].style:
        if from_global := mode_style.get("from_global"):
            if from_global not in tm.styles:
                return res.error(
                    f'global style "{from_global}" not found in {list(tm.styles)}'
                )
            mode_style = AttrDict(**tm.styles[from_global]) + mode_style

        styles.append(mode_style)

    if styles_names:
        for style in styles_names:
            if style not in tm.styles:
                return res.error(
                    f'global style "{style}" not found in {list(tm.styles)}'
                )
            styles.append(tm.styles[style])

    palette: Palette
    if palette_name:
        if palette_name in tm.palettes:
            palette = tm.palettes[palette_name]
        else:
            return res.error(f'palette "{palette_name}" not found')
    else:
        mode_palette = theme.modes[mode_name].palette
        if isinstance(mode_palette, LinkPalette):
            from_global = mode_palette.from_global

            if from_global not in tm.palettes:
                return res.error(
                    f'global style "{from_global}" not found in {list(tm.palettes)}'
                )

            palette = tm.palettes[from_global]
        else:
            palette = mode_palette

    theme = deepcopy(theme)
    styles = deepcopy(styles)
    palette = palette.copy()
    base_style = deepcopy(tm.base_style)

    theme_dict = AttrDict(palette.model_dump())

    theme_dict["theme_name"] = theme.name
    theme_dict["wallpaper"] = theme.modes[mode_name].wallpaper
    theme_dict["mode"] = mode_name

    theme_dict += base_style

    if theme.style:
        theme_dict += theme.style

    if theme.modes[mode_name].style:
        theme_dict += theme.modes[mode_name].style

    if styles:
        for s in styles:
            theme_dict += s

    theme_dict, pending = resolve_refs(theme_dict)
    while len(pending) > 0:
        c = len(pending)
        theme_dict, pending = resolve_refs(theme_dict)
        if len(pending) == c:
            break

    for p in pending:
        res.error(f'keyword reference for "{p}" not found')

    res.value = theme_dict

    res.ok = True
    return res


def valid_theme_name(name: str, themes: dict[str, Theme]) -> str:
    whitelist = "-_.() %s%s" % (string.ascii_letters, string.digits)
    char_limit = 20
    cleaned_filename = (
        unicodedata.normalize("NFKD", name).encode("ASCII", "ignore").decode()
    )
    cleaned_filename = "".join(c for c in cleaned_filename if c in whitelist)
    name = cleaned_filename[:char_limit].replace(" ", "_").lower().strip()

    tries = 1
    n = name
    while n in themes:
        n = f"{name}_{tries+1}"
        tries += 1
    return n


def import_image(wallpaper: Path, theme_dir: Path) -> Path:
    if wallpaper.parent != theme_dir and not (theme_dir / wallpaper.name).exists():
        wallpaper = files.import_image(wallpaper, theme_dir)
        log.info(f'"{wallpaper.name}" imported')
    return theme_dir / wallpaper.name



import os
import sys
from enum import Enum
from pathlib import Path


class Os(str, Enum):
    LINUX = "linux"
    WINDOWS = "windows"
    MAC = "mac"

    def __str__(self) -> str:
        return self.value


if os.environ.get("PIMP_TESTING"):
    HOME_DIR = Path("./tests/files/")
else:
    HOME_DIR = Path.home()

match sys.platform:
    case "win32":
        CLIENT_OS = Os.WINDOWS
        CONFIG_DIR = HOME_DIR / "AppData/Roaming"
        PIMP_CONFIG_DIR = HOME_DIR / "pimpmyrice"
    case "linux":
        CLIENT_OS = Os.LINUX
        CONFIG_DIR = HOME_DIR / ".config"
        PIMP_CONFIG_DIR = CONFIG_DIR / "pimpmyrice"
    case "darwin":
        CLIENT_OS = Os.MAC
        CONFIG_DIR = HOME_DIR / "Library/Application Support"
        PIMP_CONFIG_DIR = HOME_DIR / "pimpmyrice"


PIMP_DIR = Path(os.path.abspath(os.path.join(__file__, "../../")))
VENV_DIR = PIMP_CONFIG_DIR / "venv"

if CLIENT_OS == Os.WINDOWS:
    VENV_PIP_PATH = VENV_DIR / "bin" / "pip.exe"
else:
    VENV_PIP_PATH = VENV_DIR / "bin" / "pip"

LOCAL_DIR = HOME_DIR / ".local/share"
CONFIG_FILE = PIMP_CONFIG_DIR / "config.json"
BASE_STYLE_FILE = PIMP_CONFIG_DIR / "base_style.json"
THEMES_DIR = PIMP_CONFIG_DIR / "themes"
STYLES_DIR = PIMP_CONFIG_DIR / "styles"
PALETTES_DIR = PIMP_CONFIG_DIR / "palettes"
MODULES_DIR = PIMP_CONFIG_DIR / "modules"
TEMP_DIR = PIMP_CONFIG_DIR / ".tmp/"
LOCK_FILE = TEMP_DIR / "pimpmyrice.lock"
CORE_PID_FILE = TEMP_DIR / "core.pid"
SERVER_PID_FILE = TEMP_DIR / "server.pid"
LOG_FILE = PIMP_CONFIG_DIR / "pimpmyrice.log"
REPOS_LIST = PIMP_CONFIG_DIR / "remote_repos" / "list.txt"
JSON_SCHEMA_DIR = PIMP_CONFIG_DIR / ".json_schemas"

REPOS_BASE_ADDR = "https://github.com/pimpmyrice-modules"



from pathlib import Path
from typing import Any, Union

from pimpmyrice import files
from pimpmyrice.colors import GlobalPalette, Palette
from pimpmyrice.config import CLIENT_OS, MODULES_DIR, Os
from pimpmyrice.logger import get_logger
from pimpmyrice.module_utils import (
    FileAction,
    IfRunningAction,
    LinkAction,
    Module,
    ModuleInit,
    ModuleRun,
    PythonAction,
    ShellAction,
)
from pimpmyrice.theme_utils import Style, Theme, Wallpaper
from pimpmyrice.utils import Result, parse_string_vars

log = get_logger(__name__)


# TODO use pydantic


def parse_wallpaper(
    wallpaper: Union[dict[str, Any], str], theme_path: Path
) -> Wallpaper:
    match wallpaper:
        case str(wallpaper):
            return Wallpaper(path=theme_path / wallpaper)
        case dict(wallpaper):
            return Wallpaper(**{**wallpaper, "path": theme_path / wallpaper["path"]})
        case _:
            raise Exception('"wallpaper" must be a string or a dict')


def parse_theme(
    path: Path,
    global_styles: dict[str, Style],
    global_palettes: dict[str, GlobalPalette],
) -> Theme:
    name = path.name

    data = files.load_json(path / "theme.json")

    data["wallpaper"] = parse_wallpaper(data["wallpaper"], path)

    modes = data.get("modes")
    if isinstance(modes, dict):
        for mode_name, mode in modes.items():
            mode["name"] = mode_name
            if isinstance(mode, dict):
                if "wallpaper" not in mode:
                    mode["wallpaper"] = data.get("wallpaper")
                else:
                    mode["wallpaper"] = parse_wallpaper(mode["wallpaper"], path)

    theme = Theme(**data, name=name, path=path)
    # TODO global style
    return theme


def parse_module(module_path: Path) -> Result[Module]:
    res: Result[Module] = Result()

    module_name = module_path.name
    module_yaml = module_path / "module.yaml"
    module_json = module_path / "module.json"

    try:
        if module_yaml.exists():
            data = files.load_yaml(module_yaml)
        elif module_json.exists():
            data = files.load_json(module_json)
        else:
            return res

        for param in ["init", "pre_run", "run"]:
            for action in data.get(param, []):
                if isinstance(action, dict):
                    action["module_name"] = module_name

        for cmd_name, cmd in data.get("commands", {}).items():
            cmd["module_name"] = module_name

        module = Module(**data, name=module_name)

        res.value = module

    except Exception as e:
        res.exception(e)
        res.error(f'failed loading module in "{module_path}": {e}')
    finally:
        return res



from __future__ import annotations

import colorsys
from collections import Counter
from pathlib import Path
from typing import Any, Literal, Tuple

from pydantic import BaseModel, Field, model_validator
from pydantic.json_schema import SkipJsonSchema
from pydantic_core import PydanticCustomError
from pydantic_extra_types.color import RGBA
from pydantic_extra_types.color import Color as PydanticColor
from pydantic_extra_types.color import ColorType, parse_str, parse_tuple

from pimpmyrice import files
from pimpmyrice.config import PALETTES_DIR
from pimpmyrice.logger import get_logger
from pimpmyrice.utils import Timer

log = get_logger(__name__)


class Color(PydanticColor):
    hue: int = 180
    saturation: int = 50
    lightness: int = 50

    def __init__(self, value: ColorType) -> None:
        self._rgba: RGBA
        self._original: ColorType
        if isinstance(value, (tuple, list)):
            self._rgba = parse_tuple(value)
        elif isinstance(value, str):
            self._rgba = parse_str(value)
        elif isinstance(value, Color):
            self._rgba = value._rgba
            value = value._original
        else:
            raise PydanticCustomError(
                "color_error",
                "value is not a valid color: value must be a tuple, list or string",
            )

        self._original = value

        rgb = self.as_rgb_tuple()
        h, l, s = colorsys.rgb_to_hls(*tuple(x / 255 for x in rgb))

        self.hue, self.saturation, self.lightness = (
            int(h * 360),
            int(s * 100),
            int(l * 100),
        )

    @property
    def rgb(self) -> tuple[int, int, int]:
        rgb = self.as_rgb_tuple(alpha=False)[:3]
        return rgb

    @property
    def rgb_string(self) -> str:
        rgb = self.as_rgb()
        return rgb

    @property
    def hsl(self) -> tuple[int, int, int]:
        return self.hue, self.saturation, self.lightness

    @property
    def alt(self) -> Color:
        h, s, v = colorsys.rgb_to_hsv(*[x / 255 for x in self.as_rgb_tuple()])
        if v > 0.5:
            v -= 0.1
        else:
            v += 0.1
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        rgb = tuple(int(x * 255) for x in (r, g, b))
        clr = Color(rgb)  # type:ignore
        return clr

    @property
    def maxsat(self) -> Color:
        h, *_ = self.as_hsl_tuple()
        hsl = f"hsl({int(h*360)}, 100%, 50%)"
        clr = Color(hsl)

        return clr

    @property
    def hex(self) -> str:
        rgb = tuple(f"{hex(int(x))[2:] :0>2}" for x in self.as_rgb_tuple())
        hex_string = "#" + "".join(rgb)

        return hex_string

    @property
    def nohash(self) -> str:
        hex = self.hex
        clr = hex[1:]
        return clr

    @property
    def hsv(self) -> tuple[int, float, float]:
        rgb = self.as_rgb_tuple()
        h, s, v = colorsys.rgb_to_hsv(*[x / 255 for x in rgb])
        clr = int(h * 360), s, v
        return clr

    # TODO hsl
    # @property
    # def hsl(self) -> tuple[int, float, float]:
    #     return clr

    def __str__(self) -> str:
        return self.hex


class TermColors(BaseModel):
    color0: Color
    color1: Color
    color2: Color
    color3: Color
    color4: Color
    color5: Color
    color6: Color
    color7: Color
    color8: Color
    color9: Color
    color10: Color
    color11: Color
    color12: Color
    color13: Color
    color14: Color
    color15: Color


class BgFgColors(BaseModel):
    bg: Color
    fg: Color


class Palette(BaseModel):
    term: TermColors
    normal: BgFgColors
    panel: BgFgColors
    dialog: BgFgColors
    input: BgFgColors
    muted: BgFgColors
    primary: BgFgColors
    secondary: BgFgColors
    accent: BgFgColors
    destructive: BgFgColors
    border: dict[Literal["active"] | Literal["inactive"], Color]


class LinkPalette(BaseModel):
    from_global: str


class GlobalPalette(Palette):
    name: str
    path: SkipJsonSchema[Path | None] = Field(default=None, exclude=True)


def get_palettes() -> dict[str, GlobalPalette]:
    palettes = {}
    for file in PALETTES_DIR.iterdir():
        try:
            palette = files.load_json(file)
            palettes[file.stem] = GlobalPalette(name=file.stem, path=file, **palette)
        except Exception as e:
            log.exception(e)
            log.error(f'Failed to load palette "{file.stem}"')
    return palettes


def palette_display_string(colors: Any) -> str:
    circles = []
    for i in range(16):
        circles.append(f"[{Color(colors[f'color{i}']).hex}]🔘[/]")

    palette_string = " ".join(circles[0:8]) + "\r\n" + " ".join(circles[8:])

    return palette_string


def exp_extract_colors(img: Path) -> list[tuple[tuple[int, float, float], int]]:
    import cv2
    from sklearn.cluster import KMeans

    def preprocess(raw: Any) -> Any:
        image = cv2.resize(raw, (600, 600), interpolation=cv2.INTER_AREA)
        image = image.reshape(image.shape[0] * image.shape[1], 3)
        return image

    def analyze(img: Any) -> list[tuple[tuple[int, float, float], int]]:
        clf = KMeans(n_clusters=5, random_state=0, n_init="auto")
        color_labels = clf.fit_predict(img)
        center_colors = clf.cluster_centers_
        counts = Counter(color_labels)
        ordered_colors = [center_colors[i] for i in counts.keys()]

        color_objects: list[tuple[Any, int]] = []
        for i, count in counts.items():
            b, g, r = ordered_colors[i]
            c = tuple(
                Color(
                    (
                        int(r),
                        int(g),
                        int(b),
                    )
                ).hsv,
            )

            color_objects.append((c, count // 100))

        color_objects.sort(key=lambda x: x[1], reverse=True)

        return color_objects

    image = cv2.imread(str(img))
    # image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    modified_image = preprocess(image)
    colors = analyze(modified_image)

    return colors


async def exp_gen_palette(img: Path, light: bool = False) -> Palette:
    # TODO refactor everything
    # hsv: 340.32, 0.9, 0.9

    def apply_rule(clr: Tuple[float, ...], rule: dict[str, float]) -> Tuple[float, ...]:
        h, s, v = clr

        if s < rule["min_sat"]:
            s = rule["min_sat"]
        elif s > rule["max_sat"]:
            s = rule["max_sat"]

        if v < rule["min_val"]:
            v = rule["min_val"]
        elif v > rule["max_val"]:
            v = rule["max_val"]

        return h, s, v

    def are_hues_close(h1: float, h2: float, r: int = 30) -> bool:
        if abs(h1 - h2) < r:
            return True
        elif h1 - r < 0 and h1 + 360 - h2 < r:
            return True
        elif h2 - r < 0 and h2 + 360 - h1 < r:
            return True
        return False

    timer = Timer()

    extracted_hsv_colors = exp_extract_colors(img)

    main_color = extracted_hsv_colors[0][0]

    by_sat = sorted(extracted_hsv_colors, reverse=True, key=lambda x: x[0][1])
    by_sat_no_dark = [c for c in by_sat if c[0][2] > 0.3]

    saturated_colors = [
        c[0] for c in sorted(by_sat_no_dark, reverse=True, key=lambda x: x[1])
    ]

    if are_hues_close(saturated_colors[0][0], main_color[0]):
        saturated_colors.pop(0)

    primary = saturated_colors[0] if len(saturated_colors) > 0 else by_sat.pop(0)[0]
    secondary = saturated_colors[1] if len(saturated_colors) > 1 else by_sat.pop(0)[0]
    accent = saturated_colors[2] if len(saturated_colors) > 2 else by_sat.pop(0)[0]

    most_saturated = primary

    dark_rules: dict[str, Any] = {
        "normal": {
            "color": main_color,
            "bg": {"min_sat": 0.1, "max_sat": 0.5, "min_val": 0, "max_val": 0.1},
            "fg": {"min_sat": 0.1, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "panel": {
            "color": main_color,
            "bg": {"min_sat": 0.1, "max_sat": 0.5, "min_val": 0.15, "max_val": 0.2},
            "fg": {"min_sat": 0.1, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "dialog": {
            "color": main_color,
            "bg": {"min_sat": 0.1, "max_sat": 0.5, "min_val": 0.25, "max_val": 0.35},
            "fg": {"min_sat": 0.1, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "primary": {
            "color": primary,
            "bg": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.4, "max_val": 0.9},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "secondary": {
            "color": secondary,
            "bg": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.4, "max_val": 1},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "term": {"min_sat": 0.2, "max_sat": 0.45, "min_val": 0.8, "max_val": 1},
        "input": {
            "color": primary,
            "bg": {"min_sat": 0.1, "max_sat": 0.5, "min_val": 0.25, "max_val": 0.35},
            "fg": {"min_sat": 0.1, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "border": {
            "color": primary,
            "active": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.5, "max_val": 0.9},
            "inactive": {"min_sat": 0, "max_sat": 0.3, "min_val": 0, "max_val": 0.2},
        },
        "accent": {
            "color": accent,
            "bg": {"min_sat": 0.2, "max_sat": 0.45, "min_val": 0.8, "max_val": 1},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "muted": {
            "color": main_color,
            "bg": {"min_sat": 0.1, "max_sat": 0.3, "min_val": 0.8, "max_val": 1},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
        "destructive": {
            "color": accent,
            "bg": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.4, "max_val": 0.9},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 1},
        },
    }

    light_rules: dict[str, Any] = {
        "normal": {
            "color": main_color,
            "bg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.9, "max_val": 0.95},
            "fg": {"min_sat": 0, "max_sat": 0.3, "min_val": 0, "max_val": 0.12},
        },
        "panel": {
            "color": main_color,
            "bg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.85, "max_val": 0.9},
            "fg": {"min_sat": 0, "max_sat": 0.3, "min_val": 0, "max_val": 0.12},
        },
        "dialog": {
            "color": main_color,
            "bg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.7, "max_val": 0.8},
            "fg": {"min_sat": 0, "max_sat": 0.3, "min_val": 0, "max_val": 0.12},
        },
        "primary": {
            "color": primary,
            "bg": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.4, "max_val": 0.9},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 0.6},
        },
        "secondary": {
            "color": secondary,
            "bg": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.4, "max_val": 0.9},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 0.6},
        },
        "term": {"min_sat": 0.55, "max_sat": 0.8, "min_val": 0.3, "max_val": 0.5},
        "input": {
            "color": primary,
            "bg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.7, "max_val": 0.8},
            "fg": {"min_sat": 0, "max_sat": 0.3, "min_val": 0, "max_val": 0.12},
        },
        "border": {
            "color": primary,
            "active": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.6, "max_val": 0.9},
            "inactive": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.9, "max_val": 0.9},
        },
        "accent": {
            "color": accent,
            "bg": {"min_sat": 0.55, "max_sat": 0.8, "min_val": 0.45, "max_val": 0.65},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 0.6},
        },
        "muted": {
            "color": main_color,
            "bg": {"min_sat": 0.3, "max_sat": 0.5, "min_val": 0.45, "max_val": 0.65},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 0.6},
        },
        "destructive": {
            "color": accent,
            "bg": {"min_sat": 0.4, "max_sat": 0.7, "min_val": 0.4, "max_val": 0.9},
            "fg": {"min_sat": 0, "max_sat": 0.1, "min_val": 0.8, "max_val": 0.6},
        },
    }

    rules = light_rules if light else dark_rules

    palette: dict[str, Any] = {}

    for outer_name, outer in rules.items():
        if outer_name == "term":
            palette["term"] = {
                "color0": apply_rule(main_color, rules["normal"]["bg"]),
                "color15": apply_rule(main_color, rules["normal"]["fg"]),
            }
            continue
        if outer_name not in palette:
            palette[outer_name] = {}
        for inner_name, rule in outer.items():
            if inner_name == "color":
                continue
            palette[outer_name][inner_name] = apply_rule(outer["color"], rule)

    for k, v in palette.items():
        if "bg" in v and "fg" in v:
            if v["bg"][2] > 0.7 and v["bg"][2] - v["fg"][2] < 0.65:
                new_fg_v = v["bg"][2] - 0.65
                if new_fg_v < 0.1:
                    new_fg_v = 0.1
                v["fg"] = (v["fg"][0], v["fg"][1], new_fg_v)
            elif v["bg"][2] < 0.7 and v["fg"][2] - v["bg"][2] < 0.65:
                new_fg_v = v["bg"][2] + 0.65
                if new_fg_v > 0.9:
                    new_fg_v = 0.9
                v["fg"] = (v["fg"][0], v["fg"][1], new_fg_v)

    base = palette["term"]["color0"]
    palette["term"][f"color{8}"] = apply_rule(
        (base[0], 0.3, base[2] + (-0.3 if light else 0.3)), rules["term"]
    )

    for i in range(1, 8):
        hue = most_saturated[0] + 45 * (i - 1)
        if hue > 360:
            hue -= 360

        palette["term"][f"color{i}"] = apply_rule(
            (hue, most_saturated[1], most_saturated[2]), rules["term"]
        )

    for i in range(9, 15):
        h, s, v = palette["term"][f"color{i-8}"]

        palette["term"][f"color{i}"] = apply_rule(
            (
                h,
                s - 0.1,
                v - 0.1,
            ),
            rules["term"],
        )

    def to_color(d: dict[str, Any]) -> dict[str, Any]:
        for k, v in d.items():
            if isinstance(v, dict):
                d[k] = to_color(v)
            else:
                r, g, b = [
                    int(x * 255) for x in colorsys.hsv_to_rgb(v[0] / 360, v[1], v[2])
                ]
                d[k] = Color((r, g, b))

        return d

    palette = to_color(palette)

    p = Palette(**palette)

    log.info(
        f'{"light" if light else "dark"} colors for "{img.name}" generated in {timer.elapsed():.2f} seconds'
    )

    return p



import venv

from pimpmyrice.config import VENV_DIR, VENV_PIP_PATH
from pimpmyrice.logger import get_logger
from pimpmyrice.module_utils import run_shell_command

log = get_logger(__name__)


def create_venv() -> None:
    log.info("creating venv")
    venv.EnvBuilder(with_pip=True).create(VENV_DIR)
    log.success("venv created")


async def install_in_venv(packages: list[str]) -> None:
    cmd = f'{VENV_PIP_PATH} install {",".join(packages)}'
    res, err = await run_shell_command(cmd)
    print(res)
    print(err)



"""
See https://pimpmyrice.vercel.app/docs for more info.

Usage:
    pimp gen IMAGE...       [--apply] [--name=NAME] [--tags=TAGS] [options]
    pimp random             [--mode=MODE] [--name=NAME] [--tags=TAGS]
                            [--exclude-tags=TAGS] [--modules=MODULES]
                            [--exclude-modules=MODULES] [--style=STYLE]
                            [--palette=PALETTE] [--print-theme-dict]
                            [options]
    pimp refresh            [--mode=MODE] [--modules=MODULES]
                            [--exclude-modules=MODULES] [--style=STYLE]
                            [--palette=PALETTE] [--print-theme-dict]
                            [options]
    pimp set theme THEME    [--mode=MODE] [--modules=MODULES]
                            [--exclude-modules=MODULES] [--style=STYLE]
                            [--palette=PALETTE] [--print-theme-dict]
                            [options]
    pimp set mode MODE [options]
    pimp delete theme THEME [options]
    pimp rename theme THEME NEW_NAME [options]
    pimp toggle mode [options]
    pimp add tags THEMES... --tags=TAGS
    pimp remove tags [THEMES...] --tags=TAGS
    pimp clone module MODULE_URL... [options]
    pimp delete module MODULE [options]
    pimp reinit module MODULE [options]
    pimp run module MODULE COMMAND [COMMAND_ARGS...] [options]
    pimp list (themes|tags|styles|palettes|keywords|modules) [options]
    pimp edit theme [THEME] [options]
    pimp edit base-style [options]
    pimp edit style STYLE [options]
    pimp edit palette PALETTE [options]
    pimp edit module MODULE [options]
    pimp regen [--name=NAME] [options]
    pimp rewrite (themes|modules) [--name=NAME] [options]
    pimp export theme THEME OUT_DIR     [--mode=MODE] [--modules=MODULES]
                                        [--exclude-modules=MODULES] [--style=STYLE]
                                        [--palette=PALETTE] [--print-theme-dict]
                                        [options]
    pimp info [options]
    pimp --help

Options:
    --verbose, -v
"""



from __future__ import annotations

import asyncio
import os
import shutil
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pimpmyrice import module_utils as mutils
from pimpmyrice.config import LOCK_FILE, MODULES_DIR, REPOS_BASE_ADDR
from pimpmyrice.files import load_yaml, save_json, save_yaml
from pimpmyrice.logger import get_logger
from pimpmyrice.module_utils import FileAction, Module
from pimpmyrice.parsers import parse_module
from pimpmyrice.utils import AttrDict, Lock, Result, Timer, is_locked, parse_string_vars

if TYPE_CHECKING:
    from pimpmyrice.theme import ThemeManager

log = get_logger(__name__)


class ModuleManager:
    def __init__(self) -> None:
        self.modules: dict[str, Module] = {}
        self.load_modules()

    def load_modules(self) -> dict[str, Module]:
        modules: dict[str, Module] = {}
        timer = Timer()

        for module_dir in MODULES_DIR.iterdir():
            self.load_module(module_dir)

        log.debug(f"{len(modules)} modules loaded in {timer.elapsed():.4f} sec")

        return modules

    def load_module(self, module_dir: Path) -> Module | None:
        module_res = parse_module(module_dir)

        if module_res.value:
            self.modules[module_res.value.name] = module_res.value
            log.debug(f'module "{module_res.value.name}" loaded')
            return module_res.value
        return None

    async def run_modules(
        self,
        theme_dict: AttrDict,
        include_modules: list[str] | None = None,
        exclude_modules: list[str] | None = None,
        out_dir: Path | None = None,
    ) -> Result[set[str]]:

        res: Result[set[str]] = Result()

        executed_modules: set[str] = set()

        timer = Timer()

        for m in [*(include_modules or []), *(exclude_modules or [])]:
            if m not in self.modules:
                return res.error(f'module "{m}" not found')

        if is_locked(LOCK_FILE)[0]:
            return res.error("another instance is applying a theme!")

        with Lock(LOCK_FILE):
            runners = []
            pre_runners = []
            if include_modules:
                modules = {
                    name: self.modules[name]
                    for name in include_modules
                    if name in self.modules
                }
            elif exclude_modules:
                modules = {
                    name: self.modules[name]
                    for name in self.modules
                    if name not in exclude_modules
                }
            else:
                modules = self.modules

            for name, module in modules.items():
                if not module.enabled:
                    continue

                if module.pre_run:
                    pre_runners.append(name)

                if module.run:
                    runners.append(name)

            if len(runners) == 0:
                res.error(
                    "no modules to run!\n"
                    f"See {REPOS_BASE_ADDR} for available modules"
                )
                return res

            for name in pre_runners:
                mod_timer = Timer()

                mod_res = await self.modules[name].execute_pre_run(deepcopy(theme_dict))
                res += mod_res

                if mod_res.value:
                    theme_dict = mod_res.value
                    res.info(
                        f"modifier applied in {mod_timer.elapsed():.2f} seconds", name
                    )
                    executed_modules.add(name)

            modules_state = {m: {"done": False} for m in self.modules}

            tasks = [
                self.modules[name].execute_run(
                    theme_dict, modules_state=modules_state, out_dir=out_dir
                )
                for name in runners
            ]

            for t in asyncio.as_completed(tasks):
                task_res = await t
                if isinstance(task_res, Result):
                    if task_res.name:
                        executed_modules.add(task_res.name)
                        modules_state[task_res.name]["done"] = True
                    else:
                        task_res.error(f"Result has no name")

                    module_res = task_res
                else:
                    module_res = Result(name="how did this happen?")
                    module_res.exception(task_res)

                res += module_res

            res.info(
                f"{len({*pre_runners, *runners})} modules applied in {timer.elapsed():.2f} sec"
            )

            res.value = executed_modules

            res.ok = True
            return res

    async def run_module_command(
        self,
        tm: ThemeManager,
        module_name: str,
        command: str,
        *args: Any,
        **kwargs: Any,
    ) -> Result:
        res = Result()
        if module_name not in self.modules:
            return res.error(f'module "{module_name}" not found')

        module = self.modules[module_name]
        res += await module.execute_command(command, tm, *args, **kwargs)

        res.ok = True
        return res

    async def rewrite_modules(
        self,
        name_includes: str | None = None,
    ) -> Result:
        res = Result()

        for module in self.modules.values():
            if name_includes and name_includes not in module.name:
                continue

            try:
                dump = module.model_dump(mode="json")

                edited = {}
                for group_name, group in dump.items():
                    if group_name == "run":
                        for action in group:
                            if "template" in action:
                                if len(Path(action["template"]).parts) > 1:
                                    continue

                                template_path = Path(action["template"]).name
                                target_stem = Path(action["target"]).name
                                if target_stem + ".j2" == template_path:
                                    action.pop("template")

                    elif isinstance(group, (dict, list)) and len(group) == 0:
                        continue
                    elif (
                        isinstance(group, bool)
                        and group_name == "enabled"
                        and group == True
                    ):
                        continue
                    edited[group_name] = group

                save_yaml(MODULES_DIR / module.name / "module.yaml", edited)
                # save_json(MODULES_DIR / module.name / "module.json", edited)
                res.success(f'module "{module.name}" rewritten')
            except Exception as e:
                res.exception(e)
        return res

    async def init_module(self, module_name: str) -> Result:
        res = Result()

        if not module_name in self.modules:
            return res.error(f'module "{module_name}" not found')

        module = self.modules[module_name]
        init_res = await module.execute_init()
        res += init_res

        res.ok = True
        return res

    async def clone_module(self, source: str | Path | list[str | Path]) -> Result:
        res = Result()

        sources = source if isinstance(source, list) else [source]

        for source in sources:
            try:
                if isinstance(source, Path):
                    name = await mutils.clone_from_folder(source)
                elif source.startswith("pimp://"):
                    url = f'{REPOS_BASE_ADDR}/{source.removeprefix("pimp://")}'
                    name = await mutils.clone_from_git(url)
                else:
                    name = await mutils.clone_from_git(source)

                parse_res = parse_module(MODULES_DIR / name)
                res += parse_res
                if parse_res.value:
                    module = parse_res.value
                else:
                    continue

                for action in module.run:
                    if isinstance(action, FileAction):
                        target = Path(parse_string_vars(action.target))
                        if target.exists():
                            copy_path = f"{target}.bkp"
                            try:
                                shutil.copyfile(target, copy_path)
                                res.info(
                                    f'"{target.name}" copied to "{target.name}.bkp"'
                                )
                            except Exception as e:
                                res.exception(
                                    e, f'could not copy "{target}" to "{target}.bkp"'
                                )

                        link_path = Path(str(target) + ".j2")
                        if link_path.exists():
                            res.info(
                                f'skipping linking "{link_path}" to "{action.template}", destination already exists'
                            )
                            continue
                        else:
                            link_path.parent.mkdir(exist_ok=True, parents=True)

                        os.symlink(action.template, link_path)
                        res.info(f'linked "{link_path}" to "{action.template}"')

                if module.init:
                    init_res = await module.execute_init()
                    res += init_res

                # TODO clean up files
                if res.errors:
                    res.error("failed cloning module")
                    continue

                self.modules[name] = module
                res.success(f"module {name} cloned")

            except Exception as e:
                res.exception(e)
                continue

        res.ok = True
        return res

    async def delete_module(self, module_name: str) -> Result:
        res = Result()

        if module_name not in self.modules:
            return res.error(f'module "{module_name}" not found')

        module = self.modules[module_name]

        try:
            await mutils.delete_module(module)
        except Exception as e:
            res.exception(e)
        else:
            self.modules.pop(module_name)
            res.ok = True
            res.success(f'module "{module_name}" deleted')
        finally:
            return res

    async def list_modules(self) -> Result:
        res = Result()

        for module in self.modules:
            res.info(module)

        res.ok = True
        return res



import logging

from docopt import DocoptExit, docopt

from pimpmyrice.args import process_args, process_edit_args
from pimpmyrice.config import SERVER_PID_FILE
from pimpmyrice.doc import __doc__ as cli_doc
from pimpmyrice.logger import get_logger
from pimpmyrice.theme import ThemeManager
from pimpmyrice.utils import is_locked

log = get_logger(__name__)


async def cli() -> None:
    try:
        args = docopt(cli_doc)
    except DocoptExit:
        print(cli_doc)
        return

    if args["--verbose"]:
        logging.getLogger().setLevel(logging.DEBUG)

    if args["edit"]:
        await process_edit_args(args)
        return

    server_running, server_pid = is_locked(SERVER_PID_FILE)

    if server_running:
        try:
            from pimpmyrice_server.api import send_to_server
        except ImportError:
            log.error("PimpMyRice server is not installed")
            log.error("https://github.com/daddodev/pimpmyrice_server#install")
            return

        send_to_server(args)
    else:
        tm = ThemeManager()
        await process_args(tm, args)



default_base_style = {
    "bar": {"bg": "$panel.bg", "fg": "$panel.fg"},
    "blur": {"enabled": True, "passes": 3, "strength": 7},
    "border": {"radius": 0, "width": 3},
    "font": {"mono": "monospace", "normal": "sans-serif", "size": 11},
    "gaps": {"inner": 10, "outer": 20},
    "opacity": {"active": 0.98, "inactive": 0.95, "terminal": 0.97},
    "padding": {"x": 10, "y": 10},
    "shadow": {"enabled": True, "offset": 10, "opacity": 0.4, "blur": 10, "spread": 10},
    "titlebar": {
        "enabled": True,
        "active": {"bg": "$border.active", "fg": "$primary.fg"},
        "inactive": {"bg": "$border.inactive", "fg": "$normal.fg"},
    },
    "animations": {"enabled": True, "speed": 2, "style": "default"},
    "modules_styles": {},
}



from __future__ import annotations

import asyncio
import importlib.util
import os
import shlex
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Literal, Union
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field, model_validator, validator
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Annotated

from pimpmyrice import files, utils
from pimpmyrice.config import CLIENT_OS, HOME_DIR, MODULES_DIR, TEMP_DIR, Os
from pimpmyrice.logger import get_logger
from pimpmyrice.utils import AttrDict, Result, Timer, parse_string_vars

if TYPE_CHECKING:
    from pimpmyrice.theme import ThemeManager

log = get_logger(__name__)


def add_action_type_to_schema(
    action_type: str,
    schema: dict[str, Any],
) -> None:
    schema["properties"]["action"] = {
        "title": "Action type",
        "type": "string",
        "const": action_type,
    }
    schema["required"].append("action")


class ShellAction(BaseModel):
    action: Literal["shell"] = Field(default="shell")
    module_name: SkipJsonSchema[str] = Field(exclude=True)
    command: str
    detached: bool = False

    model_config = ConfigDict(
        json_schema_extra=partial(add_action_type_to_schema, "shell")
    )

    async def run(self, theme_dict: AttrDict) -> Result:
        res = Result()

        try:
            cmd = utils.parse_string_vars(
                string=self.command,
                module_name=self.module_name,
                theme_dict=theme_dict,
            )

            if self.detached:
                run_shell_command_detached(cmd)
                return res.debug(
                    f'command "{cmd}" started in background', self.module_name
                )

            out, err = await run_shell_command(cmd)

            if out:
                res.debug(out, self.module_name)
            if err:
                res.warning(
                    f'command "{cmd}" returned an error:\n{err}', self.module_name
                )

            res.debug(
                f'executed "{cmd}"',
                self.module_name,
            )
            res.ok = True
        except Exception as e:
            res.exception(e, self.module_name)
        finally:
            return res


class FileAction(BaseModel):
    action: Literal["file"] = Field(default="file")
    module_name: SkipJsonSchema[str] = Field(exclude=True)
    target: str
    template: str = ""

    model_config = ConfigDict(
        json_schema_extra=partial(add_action_type_to_schema, "file")
    )

    @model_validator(mode="before")
    @classmethod
    def set_fields(cls, data: Any) -> Any:
        if "target" in data and "template" not in data:
            template_path = f'{Path(data["target"]).name}.j2'
            data["template"] = template_path
        return data

    async def run(self, theme_dict: AttrDict, out_dir: Path | None = None) -> Result:
        res = Result()

        try:
            template = Path(
                utils.parse_string_vars(
                    string=str(
                        MODULES_DIR / self.module_name / "templates" / self.template
                    ),
                    module_name=self.module_name,
                    theme_dict=theme_dict,
                )
            )
            target = Path(
                utils.parse_string_vars(
                    string=self.target,
                    module_name=self.module_name,
                    theme_dict=theme_dict,
                )
            )

            if out_dir:
                if target.is_relative_to(HOME_DIR):
                    target = out_dir / target.relative_to(HOME_DIR)
                else:
                    target = out_dir / target

            if not target.parent.exists():
                target.parent.mkdir(parents=True, exist_ok=True)

            with open(template, "r") as f:
                data = f.read()
            processed_data = utils.process_template(data, theme_dict)

            with open(target, "w") as f:
                f.write(processed_data)

            res.debug(
                f'generated "{target.name}"',
                self.module_name,
            )
            res.ok = True
        except Exception as e:
            res.exception(e, self.module_name)
        finally:
            return res


class PythonAction(BaseModel):
    action: Literal["python"] = Field(default="python")
    module_name: SkipJsonSchema[str] = Field(exclude=True)
    py_file_path: str
    function_name: str

    model_config = ConfigDict(
        json_schema_extra=partial(add_action_type_to_schema, "python")
    )

    async def run(self, *args: Any, **kwargs: Any) -> Result[Any]:
        file_path = Path(self.py_file_path)

        if not file_path.is_absolute():
            file_path = MODULES_DIR / self.module_name / file_path

        res = Result()

        try:
            spec = importlib.util.spec_from_file_location(self.module_name, file_path)
            if not spec or not spec.loader:
                raise ImportError(f'could not load "{file_path}"')
            py_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(py_module)

            fn = getattr(py_module, self.function_name)

            res.debug(
                f"{file_path.name}:{self.function_name} loaded",
                self.module_name,
            )

            if asyncio.iscoroutinefunction(fn):
                res.value = await fn(*args, **kwargs)
            else:
                res.value = fn(*args, **kwargs)

            res.debug(
                f"{file_path.name}:{self.function_name} returned:\n{res.value}",
                self.module_name,
            )
            res.ok = True
        except Exception as e:
            res.exception(e, self.module_name)
        finally:
            return res


class WaitForAction(BaseModel):
    action: Literal["wait_for"] = Field(default="wait_for")
    module_name: SkipJsonSchema[str] = Field(exclude=True)
    module: str

    model_config = ConfigDict(
        json_schema_extra=partial(add_action_type_to_schema, "wait_for")
    )

    async def run(self, theme_dict: AttrDict, modules_state: dict[str, Any]) -> Result:
        res = Result()

        try:
            res.debug(f'waiting for module "{self.module}"...')
            # TODO add timeout
            while not modules_state[self.module]["done"]:
                await asyncio.sleep(0.1)
            res.debug(f'done waiting for module "{self.module}"')
            res.ok = True

        except Exception as e:
            res.exception(e, self.module_name)
        finally:
            return res

    def __str__(self) -> str:
        return f'wait for "{self.module}" to finish'


class IfRunningAction(BaseModel):
    action: Literal["if_running"] = Field(default="if_running")
    module_name: SkipJsonSchema[str] = Field(exclude=True)
    program_name: str
    should_be_running: bool = True

    model_config = ConfigDict(
        json_schema_extra=partial(add_action_type_to_schema, "if_running")
    )

    async def run(self, theme_dict: AttrDict) -> Result:
        res = Result()

        try:
            running = utils.is_process_running(self.program_name)
            if self.should_be_running:
                res.ok = running
            else:
                res.ok = not running
        except Exception as e:
            res.exception(e, self.module_name)
        finally:
            return res

    def __str__(self) -> str:
        return f'if "{self.program_name}" {"running" if self.should_be_running else "not running"}'


class LinkAction(BaseModel):
    action: Literal["link"] = Field(default="link")
    module_name: SkipJsonSchema[str] = Field(exclude=True)
    origin: str
    destination: str

    model_config = ConfigDict(
        json_schema_extra=partial(add_action_type_to_schema, "link")
    )

    async def run(self) -> Result:
        res = Result()

        origin_path = Path(parse_string_vars(self.origin, module_name=self.module_name))
        destination_path = Path(
            parse_string_vars(self.destination, module_name=self.module_name)
        )

        if not origin_path.is_absolute():
            origin_path = MODULES_DIR / self.module_name / "files" / origin_path

        if destination_path.exists():
            return res.error(
                f'cannot link destination "{destination_path}" to origin "{origin_path}", destination already exists'
            )
        try:
            destination_path.parent.mkdir(parents=True, exist_ok=True)
            os.symlink(
                origin_path,
                destination_path,
                target_is_directory=origin_path.is_dir(),
            )
            # action.destination.hardlink_to(action.origin)
            res.info(f'init: "{destination_path}" linked to "{origin_path}"')
            res.ok = True
        except Exception as e:
            res.exception(e, self.module_name)
        finally:
            return res


ModuleInit = Union[LinkAction]

ModulePreRun = Union[PythonAction]


ModuleRun = Union[ShellAction, FileAction, PythonAction, IfRunningAction, WaitForAction]

ModuleCommand = Union[PythonAction]


class Module(BaseModel):
    name: SkipJsonSchema[str] = Field(exclude=True)
    enabled: bool = True
    os: list[Os] = [o for o in Os]
    init: list[ModuleInit] = []
    pre_run: list[ModulePreRun] = []
    run: list[ModuleRun] = []
    commands: dict[str, ModuleCommand] = {}

    async def execute_command(
        self, command_name: str, tm: ThemeManager, *args: Any, **kwargs: Any
    ) -> Result:
        res = Result()

        if command_name not in self.commands:
            return res.error(
                f'command "{command_name}" not found in [{", ".join(self.commands.keys())}]'
            )

        try:
            action_res = await self.commands[command_name].run(tm=tm, *args, **kwargs)
            res += action_res
            if not action_res.ok:
                return res

        except Exception as e:
            return res.exception(
                e, f'command "{command_name}" encountered an error:', self.name
            )

        return res

    async def execute_init(self) -> Result:
        res = Result()

        for action in self.init:
            try:
                action_res = await action.run()
                res += action_res
                if not action_res.ok:
                    break

            except Exception as e:
                res.exception(e, f"{action} encountered an error:", self.name)
                break

        return res

    async def execute_pre_run(self, theme_dict: AttrDict) -> Result[AttrDict]:
        res: Result[AttrDict] = Result()

        try:
            for action in self.pre_run:
                action_res = await action.run(theme_dict)
                res += action_res
                if action_res.value:
                    theme_dict = action_res.value

        except Exception as e:
            res.exception(e, self.name)
        finally:
            res.value = theme_dict
            return res

    async def execute_run(
        self,
        theme_dict: AttrDict,
        modules_state: dict[str, Any],
        out_dir: Path | None = None,
    ) -> Result:
        res = Result(name=self.name)
        timer = Timer()

        # get_module_dict
        theme_dict = (
            theme_dict + theme_dict["modules_styles"][self.name]
            if self.name in theme_dict["modules_styles"]
            else deepcopy(theme_dict)
        )

        if out_dir:
            for action in self.run:
                if isinstance(action, FileAction):
                    try:
                        action_res = await action.run(theme_dict, out_dir=out_dir)
                    except Exception as e:
                        res.exception(e, f"{action} encountered an error:", self.name)
                        break

                    res += action_res
                    if not action_res.ok:
                        res.debug(f"interrupted because res.ok is false:\n{action_res}")
                        break
                else:
                    res.debug(f"dumping {action} not implemented, skipping")

            return res

        for action in self.run:
            try:
                if isinstance(action, WaitForAction):
                    action_res = await action.run(theme_dict, modules_state)
                else:
                    action_res = await action.run(theme_dict)

                res += action_res
                if not action_res.ok:
                    res.debug(f"interrupted because res.ok is false:\n{action_res}")
                    break

            except Exception as e:
                res.exception(e, f"{action} encountered an error:", self.name)
                break

        res.time = timer.elapsed()
        res.info(f"done in {res.time:.2f} sec", self.name)
        res.ok = True
        return res


def run_shell_command_detached(command: str, cwd: Path | None = None) -> None:
    if sys.platform == "win32":
        subprocess.Popen(
            shlex.split(command),
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
        )
        return

    subprocess.Popen(
        shlex.split(command),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        preexec_fn=os.setpgrp,
    )


async def run_shell_command(command: str, cwd: Path | None = None) -> tuple[str, str]:
    proc = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=cwd,
    )
    out, err = await proc.communicate()
    return out.decode(), err.decode()


def load_module_conf(module_name: str) -> dict[str, Any]:
    data = files.load_yaml(MODULES_DIR / module_name / "conf.yaml")
    return data


async def clone_from_folder(source: Path) -> str:
    if not (source / "module.yaml").exists():
        raise Exception(f'module not found at "{source}"')
    shutil.copytree(source, MODULES_DIR / source.name)
    return source.name


async def clone_from_git(url: str) -> str:
    name = url.split("/")[-1].removesuffix(".git")
    dest_dir = MODULES_DIR / name
    if dest_dir.exists():
        raise Exception(f'module "{name}" already present')
    random = str(uuid4())
    if CLIENT_OS == Os.WINDOWS:
        cmd = f'set GIT_TERMINAL_PROMPT=0 && git clone "{url}" {random}'
    else:
        cmd = f'GIT_TERMINAL_PROMPT=0 git clone "{url}" {random}'
    res, err = await run_shell_command(cmd, cwd=TEMP_DIR)

    if res:
        log.debug(res)

    if err:
        for line in err.split("\n"):
            if line and "Cloning into" not in line:
                if "terminal prompts disabled" in line:
                    raise Exception("repository not found")
                raise Exception(err)

    shutil.move(TEMP_DIR / random, dest_dir)

    return name


async def delete_module(module: Module) -> None:
    path = MODULES_DIR / module.name
    shutil.rmtree(path)



from __future__ import annotations

import os
import re
import subprocess
import types
from typing import TYPE_CHECKING, Any, Generator, Union

from docopt import (
    Argument,
    Command,
    Either,
    OneOrMore,
    Option,
    Optional,
    Required,
    formal_usage,
    parse_defaults,
    parse_pattern,
    printable_usage,
)

from pimpmyrice.config import HOME_DIR
from pimpmyrice.doc import __doc__ as cli_doc
from pimpmyrice.logger import get_logger

if TYPE_CHECKING:
    from pimpmyrice.theme import ThemeManager

log = get_logger(__name__)

# FROM infi.docopt_completion https://github.com/Infinidat/infi.docopt_completion

# We fill the file template with the command name and the different sections
# generated from the templates below.
# _message_next_arg: outputs the next positional argument name to the user
# with the _message function. It counts the number of elements in the 'words'
# special array that don't begin with '-' (options) and then uses the myargs
# array defined by the caller to output the correct argument name.
# We skip the first two elements in 'words' because the first is always empty and
# the second is the last keyword before the options and arguments start.
FILE_TEMPLATE = '''#compdef {0}

_message_next_arg()
{{
    argcount=0
    for word in "${{words[@][2,-1]}}"
    do
        if [[ $word != -* ]] ; then
            ((argcount++))
        fi
    done
    if [[ $argcount -le ${{#myargs[@]}} ]] ; then
        _message -r $myargs[$argcount]
        if [[ $myargs[$argcount] =~ ".*file.*" || $myargs[$argcount] =~ ".*path.*" ]] ; then
            _files
        fi
    fi
}}
{1}

_{0} "$@"'''

# this is a template of a function called by the completion system when crawling the arguments already
# typed. there is a section for every command and sub-command that the target script supports.
# the variables in the function, "state" and "line" are filled by the _arguments call.
# these variables are handled in "subcommand_switch", which is filled using the SUBCOMMAND_SWITCH_TEMPLATE template
SECTION_TEMPLATE = """
_{cmd_name} ()
{{
    local context state state_descr line
    typeset -A opt_args

    _arguments -C \\
        ':command:->command' \\{opt_list}
        {subcommand_switch}
}}
"""

# if "state" is "command", we call _values which lists the next available options. if state is "options" it means
# that there are more commands typed after the currently handled command, and in that case we use line[1] (the next
# option) to direct the completion system to the next section
# note that the options context is added to the _arguments call here, because this context is only supported when
# there are subcommands
SUBCOMMAND_SWITCH_TEMPLATE = """'*::options:->options'

    case $state in
        (command)
            local -a subcommands
            subcommands=(
{subcommand_list}
            )
            _values '{subcommand}' $subcommands
        ;;

        (options)
            case $line[1] in
{subcommand_cases}
            esac
        ;;
    esac
"""

CASE_TEMPLATE = """                {0})
                    _{1}-{0}
                ;;"""

# When there are positional arguments to the handled context, we use this tempalte.
# We output the name of the next positional argument by using the _message_next_args
# function (defined in FILE_TEMPLATE), unless the current word starts with "-" which means
# the user is trying to type an option (then we specify the available options by using
# _arguments, as in the regular SECTION_TEMPLATE)
ARG_SECTION_TEMPLATE = """
_{cmd_name} ()
{{
    local context state state_descr line
    typeset -A opt_args

    if [[ $words[$CURRENT] == -* ]] ; then
        _arguments -C \\
        ':command:->command' \\{opt_list}

    else
        myargs=({args})
        _message_next_arg
    fi
}}
"""


class CommandParams:
    """Contains command options, arguments, and subcommands.

    Options are optional arguments like "-v", "-h", etc.

    Arguments are required arguments like file paths, etc.

    Subcommands are optional keywords, like the "status" in "git status".
    Subcommands have their own CommandParams instance, so the "status" in "git status" can
    have its own options, arguments, and subcommands.

    This way, we can describe commands like "git remote add origin --fetch" with all the different
    options at each level.
    """

    def __init__(self) -> None:
        self.arguments: list[str] = []
        self.options: list[str] = []
        self.subcommands: dict[str, "CommandParams"] = {}

    def get_subcommand(self, subcommand: str) -> "CommandParams":
        return self.subcommands.setdefault(subcommand, CommandParams())

    def repr(self, indent: int) -> str:
        s = " " * indent + "cmds:\n"
        for cmd in self.subcommands:
            s += (
                " " * (indent + 4)
                + f"{cmd}:\n{self.subcommands[cmd].repr(indent + 5 + len(cmd))}\n"
            )
        s += " " * indent + f"args: {self.arguments}\n"
        s += " " * indent + f"opts: {self.options}\n"
        return s

    def __repr__(self) -> str:
        return self.repr(0)


class DocoptCompletionException(Exception):
    """Custom exception for docopt completion errors."""

    pass


def build_command_tree(pattern: Any, cmd_params: "CommandParams") -> "CommandParams":
    """
    Recursively fill in a command tree in cmd_params according to a docopt-parsed "pattern" object.
    """
    if type(pattern) in [Either, Optional, OneOrMore]:
        for child in pattern.children:
            build_command_tree(child, cmd_params)
    elif type(pattern) in [Required]:
        for child in pattern.children:
            cmd_params = build_command_tree(child, cmd_params)
    elif type(pattern) in [Option]:
        suffix = "=" if pattern.argcount else ""
        if pattern.short:
            cmd_params.options.append(pattern.short + suffix)
        if pattern.long:
            cmd_params.options.append(pattern.long + suffix)
    elif type(pattern) in [Command]:
        cmd_params = cmd_params.get_subcommand(pattern.name)
    elif type(pattern) in [Argument]:
        cmd_params.arguments.append(pattern.name)
    return cmd_params


def get_usage(cmd: str) -> str:
    """
    Runs a command with --help and extracts its usage description.

    :param cmd: Command to execute.
    :return: Usage string extracted from the command's help output.
    :raises: DocoptCompletionException if the command execution fails or returns an error code.
    """
    error_message = f"Failed to run '{cmd} --help'"
    try:
        cmd_process = subprocess.Popen(
            [cmd, "--help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    except OSError:
        raise DocoptCompletionException(f"{error_message} : command does not exist")
    usage = bytes()
    while True:
        nextline = cmd_process.stdout.readline()  # type: ignore
        if len(nextline) == 0 and cmd_process.poll() is not None:
            break
        usage += nextline
    if cmd_process.returncode != 0:
        raise DocoptCompletionException(
            f"{error_message} : command returned {cmd_process.returncode}"
        )
    return usage.decode("ascii")


def get_options_descriptions(doc: str) -> Generator[tuple[str, str], None, None]:
    """
    Extracts option descriptions from the docopt documentation string.

    :param doc: Documentation string from which to parse options.
    :return: A generator yielding tuples of (option, description).
    """

    def sanitize_line(line: str) -> str:
        return (
            line.replace("'", "'\\''").replace("[", "\\[").replace("]", "\\]").strip()
        )

    for arg in re.findall("\n  .*", doc):
        options, partition, description = arg.strip().partition("  ")
        if not partition:
            continue
        if not options.startswith("-"):
            yield options, sanitize_line(description)
            continue
        options = options.replace(",", " ")
        options = re.sub(r"=\S+", "= ", options)
        for s in options.split():
            yield s, sanitize_line(description)


def parse_params(cmd: str) -> tuple["CommandParams", dict[str, str]]:
    """
    Creates a parameter tree (CommandParams object) for the target docopt tool.
    Also parses a mapping of options to their help descriptions.

    :param cmd: The command to parse.
    :return: A tuple containing the CommandParams object and a dictionary of option descriptions.
    """
    usage = get_usage(cmd)
    options = parse_defaults(usage)
    pattern = parse_pattern(formal_usage(printable_usage(usage)), options)
    param_tree = CommandParams()
    build_command_tree(pattern, param_tree)
    return param_tree, dict(get_options_descriptions(usage))


class CompletionGenerator:
    """Completion file generator base class."""

    def _write_to_file(self, file_path: str, completion_file_content: str) -> None:
        if not os.access(os.path.dirname(file_path), os.W_OK):
            print(
                "Skipping file {file_path}, no permissions".format(file_path=file_path)
            )
            return
        try:
            with open(file_path, "w") as fd:
                fd.write(completion_file_content)
        except IOError:
            print("Failed to write {file_path}".format(file_path=file_path))
            return
        print("Completion file written to {file_path}".format(file_path=file_path))

    def get_name(self) -> str:
        raise NotImplementedError()

    def get_completion_path(self) -> str:
        raise NotImplementedError()

    def get_completion_filepath(
        self, cmd: str
    ) -> Union[str, Generator[str, None, None]]:
        raise NotImplementedError()

    def get_completion_file_content(
        self, cmd: str, param_tree: Any, option_help: dict[str, str]
    ) -> str:
        raise NotImplementedError()

    def completion_path_exists(self) -> bool:
        return os.path.exists(self.get_completion_path())

    def generate(self, cmd: str, param_tree: Any, option_help: dict[str, str]) -> None:
        completion_file_content = self.get_completion_file_content(
            cmd, param_tree, option_help
        )
        file_paths = self.get_completion_filepath(cmd)
        if not isinstance(file_paths, types.GeneratorType):
            file_paths = [file_paths]  # type: ignore
        for file_path in file_paths:
            self._write_to_file(file_path, completion_file_content)


class ZshCompletion(CompletionGenerator):
    """Base class for generating ZSH completion files"""

    def get_completion_path(self) -> str:
        return "."

    def get_completion_filepath(self, cmd: str) -> str:
        return os.path.join(self.get_completion_path(), f"_{cmd}")

    def create_opt_menu(self, opts: list[str], option_help: dict[str, str]) -> str:
        if not opts:
            return ""
        show_help = all(opt in option_help for opt in opts)

        def get_option_help(opt: str) -> str:
            if not show_help or opt not in option_help:
                return ""
            return f"[{option_help[opt]}]"

        def decorate_opt(opt: str) -> str:
            return (opt + "-") if opt.endswith("=") else opt

        return "\n" + "\n".join(
            [
                f"\t\t'({decorate_opt(opt)}){decorate_opt(opt)}{get_option_help(opt)}' \\"
                for opt in opts
            ]
        )

    def create_subcommand_cases(self, cmd_name: str, subcmds: list[str]) -> str:
        return "\n".join([CASE_TEMPLATE.format(cmd, cmd_name) for cmd in subcmds])

    def create_subcommand_list(
        self, cmd_name: str, option_help: dict[str, str], subcmds: list[str]
    ) -> str:
        def get_subcmd_help(subcmd: str) -> Optional[str]:
            for i in [0, 1]:
                subcommand_with_trail = " ".join(
                    cmd_name.replace("-", " ").split()[i:] + [subcmd]
                )
                if subcommand_with_trail in option_help:
                    return option_help[subcommand_with_trail]
            return None

        show_help = all(get_subcmd_help(subcmd) is not None for subcmd in subcmds)

        def get_help_opt(subcmd: str) -> str:
            if not show_help:
                return ""
            return f"[{get_subcmd_help(subcmd)}]"

        return "\n".join(
            [f"\t\t\t\t'{subcmd}{get_help_opt(subcmd)}'" for subcmd in subcmds]
        )

    def create_subcommand_switch(
        self, cmd_name: str, option_help: dict[str, str], subcommands: dict[str, Any]
    ) -> str:
        if len(subcommands) == 0:
            return ""
        subcommand_list = self.create_subcommand_list(
            cmd_name, option_help, list(subcommands.keys())
        )
        subcommand_cases = self.create_subcommand_cases(
            cmd_name, list(subcommands.keys())
        )
        return SUBCOMMAND_SWITCH_TEMPLATE.format(
            subcommand_list=subcommand_list,
            subcommand_cases=subcommand_cases,
            subcommand=cmd_name.replace("-", " "),
        )

    def create_args_section(self, cmd_name: str, opt_list: str, args: list[str]) -> str:
        return ARG_SECTION_TEMPLATE.format(
            cmd_name=f"{cmd_name}",
            args=" ".join(f"'{arg}'" for arg in args),
            opt_list=opt_list,
        )

    def create_section(
        self, cmd_name: str, param_tree: Any, option_help: dict[str, str]
    ) -> str:
        subcommands = param_tree.subcommands
        opts = param_tree.options
        args = param_tree.arguments
        opt_list = self.create_opt_menu(opts, option_help)
        if args:
            return self.create_args_section(cmd_name, opt_list, args)
        subcommand_switch = self.create_subcommand_switch(
            cmd_name, option_help, subcommands
        )
        res = SECTION_TEMPLATE.format(
            cmd_name=cmd_name, opt_list=opt_list, subcommand_switch=subcommand_switch
        )
        for subcommand_name, subcommand_tree in subcommands.items():
            res += self.create_section(
                f"{cmd_name}-{subcommand_name}", subcommand_tree, option_help
            )
        return res

    def get_completion_file_content(
        self, cmd: str, param_tree: Any, option_help: dict[str, str]
    ) -> str:
        completion_file_inner_content = self.create_section(
            cmd, param_tree, option_help
        )
        return FILE_TEMPLATE.format(cmd, completion_file_inner_content)


class OhMyZshCompletion(ZshCompletion):
    def get_name(self) -> str:
        return "ZSH with oh-my-zsh"

    def get_completion_path(self) -> str:
        return os.path.expanduser("~/.oh-my-zsh")

    def get_completion_filepath(self, cmd: str) -> str:
        completion_path = os.path.expanduser("~/.oh-my-zsh/completions")
        if not os.path.exists(completion_path):
            os.makedirs(completion_path)
        return os.path.join(completion_path, f"_{cmd}")


# END FROM infi.docopt_completion


def add_zsh_suggestions(file_content: str, arg_name: str, values: list[str]) -> str:
    if arg_name == "--tags":
        replaced = ""

        found_tags = False
        for i, line in enumerate(file_content.splitlines()):
            if "(--tags=-)--tags=-" in line:
                found_tags = True
                line = "		'--tags=:flag:->flags' \\"
            elif "}" in line and found_tags:
                found_tags = False
                line = f"""
    case "$state" in flags)
        _values -s , 'flags' {" ".join(f'"{x}"' for x in values)}
    esac
}}
"""

            replaced += line + "\n"

    elif arg_name == "IMAGE":

        replaced = file_content.replace(
            f"""
        myargs=('{arg_name.upper()}')
        _message_next_arg
""",
            f"""
        myargs=('{arg_name.upper()}')
        _files
""",
        )
    else:

        replaced = file_content.replace(
            f"""
        myargs=('{arg_name.upper()}')
        _message_next_arg
""",
            f"""
        local -a available_{arg_name}s
        available_{arg_name}s=({" ".join(f'"{x}"' for x in values)})

        _describe '{arg_name} name' available_{arg_name}s
""",
        )

    return replaced


def generate_shell_suggestions(tm: ThemeManager) -> None:
    file_path = HOME_DIR / ".cache/oh-my-zsh/completions/_pimp"

    file_path.parent.mkdir(parents=True, exist_ok=True)

    # TODO fork docopt_completion

    doc = cli_doc

    options = parse_defaults(doc)
    pattern = parse_pattern(formal_usage(printable_usage(doc)), options)
    param_tree = CommandParams()
    build_command_tree(pattern, param_tree)
    option_help = dict(list(get_options_descriptions(doc)))

    generator = ZshCompletion()

    content = generator.get_completion_file_content("pimp", param_tree, option_help)

    content = add_zsh_suggestions(content, "theme", [*tm.themes.keys()])
    content = add_zsh_suggestions(content, "module", [*tm.mm.modules.keys()])
    content = add_zsh_suggestions(content, "--tags", list(tm.tags))
    content = add_zsh_suggestions(content, "IMAGE", [])

    try:
        with open(file_path, "w") as fd:
            fd.write(content)
    except IOError:
        log.debug("Failed to write {file_path}".format(file_path=file_path))
        return
    log.debug("Completion file written to {file_path}".format(file_path=file_path))



import os
import pkgutil
from importlib.metadata import version

from pimpmyrice.files import check_config_dirs
from pimpmyrice.logger import set_up_logging

__all__ = [name for _, name, _ in pkgutil.iter_modules([os.path.dirname(__file__)])]
for module in __all__:
    __import__(f"{__name__}.{module}")

__version__ = version("pimpmyrice")

set_up_logging()



import asyncio

from pimpmyrice.cli import cli
from pimpmyrice.logger import get_logger

log = get_logger(__name__)


def main() -> None:
    try:
        asyncio.run(cli())
    except KeyboardInterrupt:
        log.info("stopped")


if __name__ == "__main__":
    main()
    # import yappi
    #
    # with yappi.run():
    #     main()
    #
    # stats = yappi.get_func_stats()
    # stats.print_all()
    # stats.save("cr.prof", type="callgrind")

    # import cProfile
    #
    # cProfile.run("main()", sort="cumtime")

    # print("profiled")



import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

# TODO

os.environ["PIMP_TESTING"] = "True"
from pimpmyrice.theme import ThemeManager


# Set environment variables and prepare test environment
@pytest.fixture(scope="session", autouse=True)
def setup_environment() -> Any:
    files_dir = Path("./tests/files")
    if files_dir.exists():
        shutil.rmtree(files_dir)
    yield
    shutil.rmtree(files_dir)


@pytest.fixture
def tm() -> ThemeManager:
    return ThemeManager()


@pytest.mark.asyncio(scope="session")
async def test_clone_module(tm: ThemeManager) -> None:
    res = await tm.mm.clone_module("pimp://alacritty")
    print(res)
    assert res.ok


@pytest.mark.asyncio(scope="session")
async def test_gen_theme(tm: ThemeManager) -> None:
    res = await tm.generate_theme("./tests/example.jpg")
    print(res)
    assert res.ok


@pytest.mark.asyncio(scope="session")
async def test_set_random_theme(tm: ThemeManager) -> None:
    res = await tm.set_random_theme()
    print(res)
    assert res.ok



import os
from importlib.metadata import version
from pathlib import Path
from typing import Any

from pimpmyrice.config import (
    BASE_STYLE_FILE,
    CONFIG_FILE,
    MODULES_DIR,
    PALETTES_DIR,
    STYLES_DIR,
    THEMES_DIR,
)
from pimpmyrice.files import load_json
from pimpmyrice.logger import get_logger
from pimpmyrice.theme import ThemeManager
from pimpmyrice.theme_utils import ThemeConfig
from pimpmyrice.utils import Result

log = get_logger(__name__)


async def process_edit_args(args: dict[str, Any]) -> None:
    def open_editor(dir: Path) -> None:
        os.system(f'$EDITOR "{dir}"')

    if not args["edit"]:
        return

    if args["base-style"]:
        open_editor(BASE_STYLE_FILE)
    elif args["theme"]:
        theme = args["THEME"]

        if not theme:
            config = load_json(CONFIG_FILE)
            theme = ThemeConfig(**config).theme

        theme_json_path = THEMES_DIR / theme / "theme.json"
        if not theme_json_path.is_file():
            log.error(f'theme "{theme}" not found')
            return

        open_editor(theme_json_path)

    elif args["style"]:
        style = args["STYLE"]

        style_path = STYLES_DIR / f"{style}.json"
        if not style_path.is_file():
            log.error(f'style "{style}" not found')
            return

        open_editor(style_path)

    elif args["palette"]:
        palette = args["PALETTE"]

        palette_path = PALETTES_DIR / f"{palette}.json"
        if not palette_path.is_file():
            log.error(f'palette "{palette}" not found')
            return

        open_editor(palette_path)

    elif args["module"]:
        module = args["MODULE"]

        module_path = MODULES_DIR / module
        if not (module_path / "module.yaml").is_file():
            log.error(f'module "{module}" not found')
            return

        open_editor(module_path)


async def process_args(tm: ThemeManager, args: dict[str, Any]) -> Result:
    res = Result()

    options = {
        "mode_name": args["--mode"],
        "palette_name": args["--palette"],
        "print_theme_dict": args["--print-theme-dict"],
    }

    if styles_names := args["--style"]:
        options["styles_names"] = styles_names.split(",")

    if modules := args["--modules"]:
        options["include_modules"] = modules.split(",")
    elif modules := args["--exclude-modules"]:
        options["exclude_modules"] = modules.split(",")

    if args["random"]:
        if name_includes := args["--name"]:
            options["name_includes"] = name_includes
        if tags := args["--tags"]:
            options["include_tags"] = tags.split(",")
        if tags := args["--exclude-tags"]:
            options["exclude_tags"] = tags.split(",")
        return await tm.set_random_theme(**options)

    elif args["refresh"]:
        return await tm.apply_theme(**options)

    elif args["theme"]:
        if args["set"]:
            return await tm.apply_theme(theme_name=args["THEME"], **options)

        elif args["rename"]:
            return await tm.rename_theme(
                theme_name=args["THEME"],
                new_name=args["NEW_NAME"],
            )

        elif args["delete"]:
            return tm.delete_theme(args["THEME"])
        elif args["export"]:
            return await tm.export_theme(
                args["THEME"], out_dir=Path(args["OUT_DIR"]).absolute(), **options
            )

    elif args["module"]:
        if args["clone"]:
            return await tm.mm.clone_module(args["MODULE_URL"])

        elif args["delete"]:
            return await tm.mm.delete_module(args["MODULE"])

        elif args["run"]:
            return await tm.mm.run_module_command(
                tm,
                module_name=args["MODULE"],
                command=args["COMMAND"],
                cmd_args=args["COMMAND_ARGS"],
            )

        elif args["reinit"]:
            return await tm.mm.init_module(module_name=args["MODULE"])

    elif args["tags"]:
        tags = set(args["--tags"].split(","))
        if args["add"]:
            return await tm.add_tags(args["THEMES"], tags)
        elif args["remove"]:
            return await tm.remove_tags(args["THEMES"], tags)

    elif args["toggle"]:
        return await tm.toggle_mode()

    elif args["mode"]:
        mode = args["MODE"]

        return await tm.set_mode(mode)

    elif args["gen"]:
        a = {}

        if tags := args["--tags"]:
            a["tags"] = tags.split(",")

        if args["--name"]:
            a["name"] = args["--name"]

        if apply := args["--apply"]:
            a["apply"] = apply

        for img in args["IMAGE"]:
            r = await tm.generate_theme(image=img, **a)
            res += r

        return res

    elif args["list"]:
        if args["modules"]:
            return await tm.mm.list_modules()
        elif args["themes"]:
            return await tm.list_themes()
        elif args["tags"]:
            return await tm.list_tags()
        elif args["palettes"]:
            return await tm.list_palettes()
        elif args["styles"]:
            return await tm.list_styles()

    elif args["info"]:
        # TODO use Rich Table?
        msg = f"""🍙 PimpMyRice {version("pimpmyrice")}
name: {tm.config.theme}
mode: {tm.config.mode}
"""

        return res.info(msg)

    elif args["regen"]:
        return await tm.rewrite_themes(regen_colors=True, name_includes=args["--name"])

    elif args["rewrite"]:
        if args["themes"]:
            return await tm.rewrite_themes(name_includes=args["--name"])
        elif args["modules"]:
            return await tm.mm.rewrite_modules(name_includes=args["--name"])

    return res.error("not implemented")



import setuptools

with open("README.md", "r", encoding="utf8") as f:
    long_description = f.read()

setuptools.setup(
    name="pimpmyrice",
    version="0.3.1",
    author="daddodev",
    author_email="daddodev@gmail.com",
    description="The overkill rice manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/daddodev/pimpmyrice",
    project_urls={
        "Bug Tracker": "https://github.com/daddodev/pimpmyrice/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["pimp=pimpmyrice.__main__:main"]},
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.10",
    install_requires=[
        "docopt",
        "jinja2",
        "pyyaml",
        "psutil",
        "requests",
        "scikit-learn",
        "opencv-python",
        "rich",
        "typing_extensions",
        "pydantic",
        "pydantic-extra-types",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-asyncio",
            "mypy",
            "black",
            "types-requests",
            "types-PyYAML",
            "isort",
        ],
    },
)



from functools import partial
from typing import Any, Callable, Coroutine


class EventHandler:
    def __init__(self) -> None:
        self.subscribers: dict[str, list[Callable[[], Coroutine[Any, Any, None]]]] = {}

    def subscribe(
        self, event_name: str, fn: Callable[..., Coroutine[Any, Any, None]], *args: Any
    ) -> None:
        if event_name not in self.subscribers:
            self.subscribers[event_name] = []

        self.subscribers[event_name].append(partial(fn, *args))

    async def publish(self, event_name: str) -> None:
        if event_name not in self.subscribers:
            return
        for callback in self.subscribers[event_name]:
            await callback()
